# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import sys
try:
    from cStringIO import StringIO
except:
    from io import StringIO

from byot import (
    assertions,
    fixtures,
    results,
    runners,
    scenarii,
    suites,
)


load_tests = scenarii.load_tests_with_scenarios


class TestSuiteSetup(unittest.TestCase):

    scenarios = [('unittest', dict(case_maker=fixtures.make_case)),
                 ('testtools', dict(case_maker=fixtures.make_testtools_case))]

    def run_suite(self, suite):
        res = results.TextResult(StringIO(), verbosity=0)
        res.startTestRun()
        suite.run(res)
        res.stopTestRun()
        return res

    def test_setup_pass(self):
        self.setup_called = False

        class Suite(suites.TestSuite):

            def setUp(inner):
                super(Suite, inner).setUp()
                self.setup_called = True

        suite = fixtures.make_suite(['pass', 'pass'], suite_maker=Suite)
        res = self.run_suite(suite)
        self.assertTrue(self.setup_called)
        self.assertTrue(res.wasSuccessful())
        self.assertEqual(2, res.testsRun)
        self.assertEqual(0, len(res.errors))
        self.assertEqual(0, len(res.failures))

    def test_setup_errors(self):
        self.setup_called = False

        class Suite(suites.TestSuite):

            def setUp(inner):
                super(Suite, inner).setUp()
                self.setup_called = True
                raise AssertionError

        suite = fixtures.make_suite(['pass', 'pass'], suite_maker=Suite)
        res = self.run_suite(suite)
        # An error during the setup fails the run without running the tests
        self.assertTrue(self.setup_called)
        self.assertFalse(res.wasSuccessful())
        self.assertEqual(0, res.testsRun)
        self.assertEqual(1, len(res.errors))
        self.assertEqual(0, len(res.failures))
        self.assertEqual(
            'byot.tests.test_suites.Suite({:x})'.format(id(suite)),
            res.errors[0][0].id())

    def test_tests_pass(self):
        self.setup_called = False
        self.cleanup_called = False

        class Suite(suites.TestSuite):

            def setUp(inner):
                super(Suite, inner).setUp()
                self.setup_called = True

                def cleanup():
                    self.cleanup_called = True
                inner.addCleanup(cleanup)

        suite = fixtures.make_suite(['pass', 'pass'], suite_maker=Suite)
        res = self.run_suite(suite)
        # When the tests pass, setup and cleanups are called.
        self.assertTrue(self.setup_called)
        self.assertTrue(self.cleanup_called)
        self.assertTrue(res.wasSuccessful())
        self.assertEqual(2, res.testsRun)
        self.assertEqual(0, len(res.errors))
        self.assertEqual(0, len(res.failures))

    def test_tests_errors(self):
        self.setup_called = False
        self.cleanup_called = False

        class Suite(suites.TestSuite):

            def setUp(inner):
                super(Suite, inner).setUp()
                self.setup_called = True

                def cleanup():
                    self.cleanup_called = True
                inner.addCleanup(cleanup)

        suite = fixtures.make_suite(['pass', 'error'], suite_maker=Suite)
        res = self.run_suite(suite)
        # If a test fails, the cleanups are called
        self.assertTrue(self.setup_called)
        self.assertTrue(self.cleanup_called)
        self.assertFalse(res.wasSuccessful())
        self.assertEqual(2, res.testsRun)
        self.assertEqual(1, len(res.errors))
        self.assertEqual(0, len(res.failures))

    def test_cleanup_errors(self):
        class Suite(suites.TestSuite):

            def setUp(inner):
                super(Suite, inner).setUp()

                def cleanup():
                    raise AssertionError
                inner.addCleanup(cleanup)

        suite = fixtures.make_suite(['pass', 'pass'], suite_maker=Suite)
        res = self.run_suite(suite)
        # If a cleanup fails, the suite fails even if the tests succeed
        self.assertFalse(res.wasSuccessful())
        self.assertEqual(2, res.testsRun)
        self.assertEqual(1, len(res.errors))
        self.assertEqual(0, len(res.failures))
        self.assertEqual(
            'byot.tests.test_suites.Suite({:x})'.format(id(suite)),
            res.errors[0][0].id())

invocation_path = None


class TestingSubprocessSuiteWithLoadFile(suites.SubprocessedSuiteWithLoadFile):

    def setUp(self):
        super(TestingSubprocessSuiteWithLoadFile, self).setUp()
        fixtures.override_env(self, 'BYOTSELFTESTS_IN_SUBPROCESS', '1')
        # Make sure we call the right top-level script
        self.cmd = [sys.executable, invocation_path,
                    '--module=byot',
                    '--format=subunit', '--load', self.test_list_path]


class TestForSubprocess(unittest.TestCase):
    """Special tests for subprocesses.

    We need failing and erroring tests in subprocesses for
    TestRunTestsConcurrent but we want them to succeed otherwise.

    We can't use the usual tests in fixtures because those are not loaded and
    as such can't be used with --load.

    Not all subprocess tests can rely on --load either.

    If you need to update this comment, time has come to re-think all the use
    cases and probably create a python file on the fly with whatever test is
    needed.
    """

    @staticmethod
    def make_suite(kinds):
        return fixtures.make_suite(
            kinds, lambda kind: TestForSubprocess('test_{}'.format(kind)))

    def test_pass(self):
        pass

    def test_fail(self):
        if os.environ.get('BYOTSELFTESTS_IN_SUBPROCESS', False):
            raise self.failureException

    def test_error(self):
        if os.environ.get('BYOTSELFTESTS_IN_SUBPROCESS', False):
            raise SyntaxError('invalid syntax')


class TestRunTestsConcurrent(unittest.TestCase):

    scenarios = [
        ('fork', dict(suite_class=suites.ForkedSuite,
                      suite_maker=fixtures.make_suite)),
        ('subproc', dict(suite_class=TestingSubprocessSuiteWithLoadFile,
                         suite_maker=TestForSubprocess.make_suite)),
    ]

    def setUp(self):
        super(TestRunTestsConcurrent, self).setUp()
        # Save the invocation path for tests that needs it
        global invocation_path
        invocation_path = os.path.realpath(sys.argv[0])
        # Some tests will create files (test list), guard against collisions
        fixtures.set_uniq_cwd(self)

    def run_suite(self, suite):
        res = results.TextResult(StringIO(), verbosity=0)
        # Run tests across 2 processes, it's enough to validate that the
        # plumbing is right between the main process and the subprocesses.
        runners.run_tests(suite, res, 2, split_suite=self.suite_class)
        return res

    def test_pass(self):
        res = self.run_suite(self.suite_maker(['pass', 'pass']))
        self.assertTrue(res.wasSuccessful())
        self.assertEqual(2, res.testsRun)
        assertions.assertLength(self, 0, res.errors)
        assertions.assertLength(self, 0, res.failures)

    def test_fail(self):
        res = self.run_suite(self.suite_maker(['pass', 'fail']))
        self.assertFalse(res.wasSuccessful())
        self.assertEqual(2, res.testsRun)
        assertions.assertLength(self, 0, res.errors)
        assertions.assertLength(self, 1, res.failures)

    def test_error(self):
        res = self.run_suite(self.suite_maker(['error', 'fail']))
        self.assertFalse(res.wasSuccessful())
        self.assertEqual(2, res.testsRun)
        assertions.assertLength(self, 0, res.errors)
        assertions.assertLength(self, 2, res.failures)


class TestingPickedSuite(suites.SubprocessedSuite):

    def setUp(self):
        super(TestingPickedSuite, self).setUp()
        fixtures.override_env(self, 'BYOTSELFTESTS_IN_SUBPROCESS', '1')
        # Make sure we call the right top-level script
        self.cmd = [sys.executable, sys.argv[0],
                    '--module=byot',
                    '--format=subunit',
                    '--pick', '{}/{}'.format(self.unique, self.factor),
                    # Restrict to the class we're testing so we don't recurse
                    TestForSubprocess.__name__]


class TestRunTestsConcurrentPicked(unittest.TestCase):
    """Test pick specifics.

    We can't reuse the tests from TestRunTestsConcurrent as they create test
    lists and --pick is aimed at avoiding test lists.
    """

    suite_class = TestingPickedSuite

    def run_suite(self, suite):
        res = results.TextResult(StringIO(), verbosity=0)
        # Run tests across 3 processes (because there are 3 tests ;) to
        # validate that the plumbing is right between the main process and the
        # subprocesses.
        runners.run_tests(suite, res, 3, split_suite=self.suite_class)
        res.stopTestRun()
        return res

    def test_them_all(self):
        # the subprocesses won't see the suite so we may as well provide an
        # empty one
        res = self.run_suite(fixtures.make_suite([]))
        self.assertFalse(res.wasSuccessful())
        self.assertEqual(3, res.testsRun)
        # testtools streams merge errors and failures
        assertions.assertLength(self, 0, res.errors)
        assertions.assertLength(self, 2, res.failures)


class TestRunSuiteForked(unittest.TestCase):

    def test_fork_fails(self):

        def failing_fork():
            raise OSError('os.fork() failed')

        fixtures.patch(self, os, 'fork', failing_fork)
        with self.assertRaises(OSError) as cm:
            suite = suites.ForkedSuite(unittest.TestSuite(), 0, 0)
            suite.run(results.TextResult(StringIO()))
        self.assertEqual('os.fork() failed', '{}'.format(cm.exception))

# MISSING tests: ForkedSuite, and SubprocessedSuite as well, should be tested
# without involving split_suite_for(). I.e. both parts of the subunit
# communication should be tested independently (which may be a hint that the
# current implementation is still not testable enough) -- vila 2015-12-04



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2014, 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

import byot
from byot import (
    features,
    fixtures,
    styles,
    tests,
)


@features.requires(tests.minimal_pep8)
class TestTestPep8Usage(unittest.TestCase):

    def test_no_packages(self):
        with self.assertRaises(AssertionError) as cm:

            class Pep8(styles.TestPep8):

                packages = []

            test = Pep8('test_pep8_conformance')
            test.test_pep8_conformance()

        self.assertEqual(
            '[] == [] : You should define some packages to check',
            str(cm.exception))


@features.requires(tests.minimal_pyflakes)
class TestTestPyflakesUsage(unittest.TestCase):

    def test_no_packages(self):
        with self.assertRaises(AssertionError) as cm:

            class Pyflakes(styles.TestPyflakes):

                packages = []

            test = Pyflakes('test_pyflakes_conformance')
            test.test_pyflakes_conformance()

        self.assertEqual(
            '[] == [] : You should define some packages to check',
            str(cm.exception))


class TestPythonWalker(unittest.TestCase):

    def setUp(self):
        super(TestPythonWalker, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_empty(self):
        pw = styles.PythonFileWalker('.')
        self.assertEqual([], list(pw.iter('')))

    def test_walk_from_root(self):
        fixtures.build_tree('''
file: foo.py
dir: a
# a/bar has no .py, it won't be selected
file: a/bar
file: a/bar.py
''')
        pw = styles.PythonFileWalker(os.getcwd())
        self.assertEqual(['a/bar.py', 'foo.py'],
                         list(pw.iter('')))

    def test_walk_from_dir_inside_root(self):
        fixtures.build_tree('''
file: foo.py
dir: a
file: a/bar.py
dir: a/b
file: a/b/baz.py
''')
        pw = styles.PythonFileWalker('a')
        self.assertEqual(['b/baz.py'],
                         list(pw.iter('b')))


class TestPyflakesCheckDir(unittest.TestCase):

    def setUp(self):
        super(TestPyflakesCheckDir, self).setUp()
        fixtures.set_uniq_cwd(self)
        self.paths = []

        def check_file(path, ignored):
            self.paths.append(path)

        self.check = check_file

    def assertPaths(self, expected, start='', excludes=None):
        walker = styles.PythonFileWalker('')
        if excludes is None:
            excludes = []
        styles.pyflakes_check_dir(walker, start,
                                  excludes, self.check, None)
        self.assertEqual(expected, self.paths)

    def test_empty(self):
        self.assertPaths([], '')

    def test_tree(self):
        fixtures.build_tree('''
file: foo
file: foo.py
dir: a
file: a/bar
file: a/bar.py
dir: b
dir: b/c
file: b/c/baz.py
''')
        self.assertPaths(['a/bar.py', 'b/c/baz.py', 'foo.py'], )

    def test_tree_with_excludes(self):
        fixtures.build_tree('''
file: foo
file: foo.py
dir: a
file: a/bar
file: a/bar.py
dir: b
dir: b/c
file: b/c/baz.py
''')
        self.assertPaths(['a/bar.py', 'b/c/baz.py'], excludes=['foo.py'])

    def test_tree_with_excludes_in_dir(self):
        fixtures.build_tree('''
file: foo
file: foo.py
dir: a
file: a/bar
file: a/bar.py
file: a/baz.py
dir: b
dir: b/c
file: b/c/baz.py
''')
        self.assertPaths(['a/baz.py'], start='a', excludes=['a/bar.py'])


# Style tests for byot itself


class TestPep8(styles.TestPep8):

    packages = [byot]


class TestPyflakes(styles.TestPyflakes):

    packages = [byot]



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import unittest

import subunit
import testtools
from subunit import test_results as su_results
from testtools import testresult as tt_results


class TextResult(tt_results.TextTestResult):
    """A TestResult which outputs activity to a text stream."""

    def __init__(self, stream, failfast=False, verbosity=1):
        super(TextResult, self).__init__(stream, failfast)
        self.verbose = verbosity > 1
        # Output is delayed until 'stopTest' is called. This ensures the output
        # is produced from a single place at a precise point in time (they are
        # cases where addError is called multiple times and delaying the output
        # provides a more consistent behavior).
        self.delayed_output = ''

    def startTest(self, test):
        if self.verbose:
            test_id = test.id()
            self.stream.write(test_id)
            self.stream.write(' ... ')
            self.stream.flush()
        self.start_time = self._now()
        super(TextResult, self).startTest(test)

    def convert_delta_to_float(self, a_timedelta):
        return (a_timedelta.days * 86400.0 + a_timedelta.seconds +
                a_timedelta.microseconds / 1000000.0)

    def stopTest(self, test):
        self.stream.write(self.delayed_output)
        self.delayed_output = ''
        if self.verbose:
            elapsed_time = self._now() - self.start_time
            self.stream.write(' (%.3f secs)\n'
                              % self.convert_delta_to_float(elapsed_time))
        self.stream.flush()
        super(TextResult, self).stopTest(test)

    def startTestRun(self):
        super(TextResult, self).startTestRun()
        # Activate signal handling so KeyboardInterrupt use the same internal
        # plumbing as failfast
        unittest.installHandler()
        unittest.registerResult(self)

    def stopTestRun(self):
        super(TextResult, self).stopTestRun()
        for reason, skipped in self.skip_reasons.items():
            if not reason:
                reason = 'No reason given'
            self.stream.write('{} skipped: {}\n'.format(
                len(skipped), reason))

    def addError(self, test, err=None, details=None):
        if self.verbose:
            self.delayed_output += 'ERROR'
        else:
            self.delayed_output += 'E'
        super(TextResult, self).addError(test, err, details)

    def addFailure(self, test, err=None, details=None):
        if self.verbose:
            self.delayed_output += 'FAIL'
        else:
            self.delayed_output += 'F'
        super(TextResult, self).addFailure(test, err, details)

    def addSkip(self, test, reason):
        if self.verbose:
            if not reason:
                reason_displayed = ''
            else:
                reason_displayed = ' ' + reason
            self.delayed_output += 'SKIP{}'.format(reason_displayed)
        else:
            self.delayed_output += 's'
        super(TextResult, self).addSkip(test, reason)

    def addSuccess(self, test, details=None):
        if self.verbose:
            self.delayed_output += 'OK'
        else:
            self.delayed_output += '.'
        super(TextResult, self).addSuccess(test, details)

    def addExpectedFailure(self, test, err=None):
        if self.verbose:
            self.delayed_output += 'XFAIL'
        else:
            self.delayed_output += 'x'
        super(TextResult, self).addExpectedFailure(test, err)

    def addUnexpectedSuccess(self, test):
        if self.verbose:
            self.delayed_output += 'NOTOK'
        else:
            self.delayed_output += 'u'
        super(TextResult, self).addUnexpectedSuccess(test)


class SubunitProducer(testtools.ExtendedToStreamDecorator):
    """A test result producing a subunit v2 stream.

    The subunit stream is produced from testtools stream result events.
    """

    def __init__(self, stream):
        streamer = subunit.StreamResultToBytes(stream)
        super(SubunitProducer, self).__init__(streamer)


class SubunitConsumer(object):
    """A test result consuming a subunit v2 stream.

    The subunit stream is turned into testtools stream result events.
    Non subunit output is routed independently.

    The object acts as a test case and run(final_result) needs to be called to
    process the events through ``final_result``.
    """

    def __init__(self, stream, noise=None):
        if noise is None:
            # Unless told otherwise, that's where the noise should end up.
            noise = sys.stdout
        self.stream = stream
        self.noise = noise

    def run(self, result):
        # The received stream will be split between the subunit data and the
        # noise
        router = testtools.StreamResultRouter(result)
        noise_consumer = su_results.CatFiles(self.noise)
        router.add_rule(noise_consumer, 'test_id', test_id=None)
        su_test = subunit.ByteStreamToStreamResult(
            self.stream, non_subunit_name='noise')
        return su_test.run(router)



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2014, 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""A collection of commonly used 'Features' to optionally run tests."""

from distutils import spawn
import unittest


class Feature(object):
    """An operating system Feature."""

    def __init__(self):
        self._available = None

    def available(self):
        """Is the feature available?

        :return: True if the feature is available.
        """
        if self._available is None:
            self._available = self._probe()
        return self._available

    def _probe(self):
        """Implement this method in concrete features.

        :return: True if the feature is available.
        """
        raise NotImplementedError

    def feature_name(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.feature_name()

    def __str__(self):
        return self.feature_name()


class ExecutableFeature(Feature):
    """Feature testing whether an executable of a given name is on the PATH."""

    def __init__(self, name):
        super().__init__()
        self.name = name
        self._path = None

    @property
    def path(self):
        # This is a property, so accessing path ensures _probe was called
        self.available()
        return self._path

    def _probe(self):
        self._path = spawn.find_executable(self.name)
        return self._path is not None

    def feature_name(self):
        return '{} executable'.format(self.name)


class PlatformFeature(Feature):
    """Feature testing the current platform."""

    def __init__(self):
        super(PlatformFeature, self).__init__()
        # self.distro structure may vary between distros, the daughter classes
        # handle the differences.
        try:
            # Removed in python3.8 according to
            # https://pypi.org/project/distro/
            # FIXME: We don't support python 3.8 anyway ? -- vila 2024-07-04
            import platform
            self.distro = platform.linux_distribution()
        except AttributeError:
            import distro
            self.distro = distro.linux_distribution()


class UbuntuPlatformFeature(PlatformFeature):
    """Feature testing the current ubuntu release where the tests are run."""

    def __init__(self, expected_id):
        super(UbuntuPlatformFeature, self).__init__()
        self.expected_id = expected_id
        self._name, self._version, self.actual_id = self.distro

    def _force_release(self, name, version, actual_id):
        """For tests: force values ignoring actual platform."""
        self._name = name
        self._version = version
        self.actual_id = actual_id

    def feature_name(self):
        return self.expected_id

    def _probe(self):
        return (self._name == 'Ubuntu' and self.expected_id == self.actual_id)

    @property
    def version(self):
        """The Ubuntu version (yy.mm) string.

        This property is available even when we run on a different version,
        so we *don't* call _probe().
        """
        return self._version

xenial = UbuntuPlatformFeature('xenial')
artful = UbuntuPlatformFeature('artful')
bionic = UbuntuPlatformFeature('bionic')
# Tested releases. The order and the names is what matters.
ubuntu_releases = ('xenial', 'artful', 'bionic')


class AtLeastUbuntu(UbuntuPlatformFeature):

    def _probe(self):
        try:
            expected = ubuntu_releases.index(self.expected_id)
            actual = ubuntu_releases.index(self.actual_id)
            return (self._name == 'Ubuntu' and expected <= actual)
        except ValueError:
            return False


at_least_xenial = AtLeastUbuntu('xenial')
at_least_bionic = AtLeastUbuntu('bionic')


def requires(feature):
    """A decorator to skip a test if a feature is not available.

    If the feature is not available, a test case or a test suite using this
    decorator will be skipped.
    """
    if not feature.available():
        reason = '{} is not available'.format(feature.feature_name())
        return unittest.skip(reason)

    # If the feature is available, there is no need to decorate
    return lambda obj: obj


def test_requires(test, feature):
    """Skip a test if a feature is not available.

    This function can be used in a setUp or test method.
    """
    if not feature.available():
        reason = '{} is not available'.format(feature.feature_name())
        test.skipTest(reason)
    return feature  # For convenience



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test scenarios handling.

This provides support for parametrized tests through the definition of a test
class attribute called scenarios describing which test objects should be
created.

A scenario is defined by a name and a dict.

The name is used to uniquely identify a test object by adding it to the test
id().

The dict defines a set of test object attributes that will be set to the
corresponding values.

The unusual (but etymologically correct) plural form of scenario is used for
the module name to avoid colliding with 'scenarios' that users are encouraged
to use for scenarios definitions.

"""

import copy
import functools
import unittest


def clone_test(test, new_id):
    """Clone a test giving it a new id.

    This should only be used on test objects that have just been created but
    not executed.

    :param test: The test to clone.

    :param new_id: The id to assign to it.

    :return: The cloned test.

    """
    clone = copy.copy(test)
    clone.id = lambda: new_id
    return clone


def apply_scenario(test, scenario):
    """Copy a test and apply a scenario.

    :param test: The test receiving the scenario

    :param scenario: A tuple describing the scenario.

        The first element of the tuple is the scenario key that will be
        embedded in the new test id.

        The second element is a dict containing attributes to set on the
        test.

    :return: The adapted test.

    """
    name, attrs = scenario
    # Adding the name between parens preserve unique ids
    new_id = '{}({})'.format(test.id(), name)
    new_test = clone_test(test, new_id)
    for name, value in attrs.items():
        setattr(new_test, name, value)
    return new_test


def apply_scenarios(test, scenarios):
    """Apply the scenarios to test and yield the cloned tests.

    :param test: The test to apply scenarios to.

    :param scenarios: An iterable of scenarios to apply to the test.
    """
    for scenario in scenarios:
        yield apply_scenario(test, scenario)


def multiply_suite(suite, scenarios=None):
    """Multiply suite by scenarios.

    :param suite: The suite to parameterize. It contains either tests or suites
        inheriting from ``unittest.TestSuite``.

    :param scenarios: The scenarios to apply. If None is supplied (the
        default), the test 'scenarios' attributes will be used when
        encountered.

    :returns: The parametrized test suite.

    ``suite`` is a tree of tests and suites, the returned suite respect the
    received suite layout.

    """
    multiplied_suite = suite.__class__()
    suite_scenarios = getattr(suite, 'scenarios', None)
    if scenarios is None and suite_scenarios is not None:
        # We use the scenarios from the suite from now on.
        scenarios = suite_scenarios
    for test in suite:
        if issubclass(test.__class__, unittest.TestSuite):
            # We received a suite, we multiply that suite
            multiplied = multiply_suite(test, scenarios)
        else:
            test_scenarios = getattr(test, 'scenarios', None)
            if scenarios is None and test_scenarios is not None:
                # We use the scenarios from the test from now on.
                scenarios = test_scenarios
            if scenarios is None:
                # No scenarios so the test is left untouched
                multiplied = [test]
            else:
                # Now we multiply, for real
                multiplied = list(apply_scenarios(test, scenarios))
        multiplied_suite.addTests(multiplied)
    return multiplied_suite


def load_tests_with_scenarios(loader, module_tests, pattern):
    """Multiply tests depending on their 'scenarios' attribute.

    This is a shorcut that can be assigned to 'load_tests' in any test module
    to make this automatically work across tests in the module (whether they
    specify a 'scenarios' attribute or not).
    """
    return multiply_suite(module_tests)


def multiply_scenarios(*scenarios):
    """Multiply two or more iterables of scenarios.

    :returns: A list of compound scenarios: the cross-product of all scenarios,
        with the names concatenated and the parameter dicts merged together.

    """

    def multiply(these_scenarios, those_scenarios):
        """Multiply two sets of scenarios.

        :returns: the cartesian product of the two sets of scenarios.
        """
        return [('{},{}'.format(this_name, that_name),
                 dict(list(this_dict.items()) + list(that_dict.items())))
                for this_name, this_dict in these_scenarios
                for that_name, that_dict in those_scenarios]

    return functools.reduce(multiply, map(list, scenarios))



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import sys


from byot import (
    features,
)


class MinimalTesttools(features.Feature):

    def _probe(self):
        import testtools
        return testtools.__version__ >= (1, 8, 1)


minimal_testtools = MinimalTesttools()


class MinimalPep8(features.Feature):
    # Supporting precise is just too much work, requires at least the saucy
    # version

    def _probe(self):
        import pep8
        return pep8.__version__ >= '1.7.0'


minimal_pep8 = MinimalPep8()


class MinimalPyflakes(features.Feature):
    # Supporting precise is just too much work, requires at least the saucy
    # version

    def _probe(self):
        import pyflakes
        return pyflakes.__version__ >= '1.1.0'


minimal_pyflakes = MinimalPyflakes()


class Python(object):
    """A few common features of python itself.

    A few, minor, differences have been tracked by the tests across the years,
    documenting them as a way to demonstrate that focused testing is cheap in
    the long term.

    This is mainly to help tests stay robust against minor changesby providing
    east to use variants to build text matching against "standard" outputs.
    """

    def __init__(self):
        super().__init__()
        self.version = sys.version_info
        if self.version >= (3, 11):
            # This is used to match the carets python uses to underline in
            # the backtrace
            self.backtrace_carets = '...\n'
        else:
            self.backtrace_carets = ''

        if self.version >= (3, 9):
            # The help format has evolved
            self.argparse_options = 'options'
        else:
            self.argparse_options = 'optional arguments'
        if self.version >= (3, 6):
            # The exception name raised when a module cannot be imported
            self.mnf_error = 'ModuleNotFoundError'
        else:
            self.mnf_error = 'ImportError'


python = Python()



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Assertions that can be used against any test case."""

# doctests ?? Nah, fear not, it's only to reuse OutputChecker which does a good
# job at fuzzy matching texts with ELLIPSIS.
import doctest
import unittest


def assertMultiLineAlmostEqual(test, expected, actual):
    """Assert that a template matches the actual output.

    Comparing multiple lines for strict equality can introduce strong
    constraints for little added value. There are cases where some part of the
    lines are not relevant to the test goal (dates, times, directories
    containing the file names compared are a few often encountered cases).

    This assertion provides a way to mark such parts in the expected string
    with ellipsis (...).

    :param test: The calling test.

    :param expected: The expected string, all ellipsis there won't produce
        differences during the comparison.

    :param actual: The actual string to compare against.
    """
    checker = doctest.OutputChecker()
    options = doctest.ELLIPSIS
    matching = checker.check_output(expected, actual, options)
    if not matching:
        # We can't use output_checker.output_difference() here because... the
        # API is broken ('expected' must be a doctest specific object of which
        # a 'want' attribute will be our 'expected' parameter). So we just
        # fallback to our assertMultilineEqual since we know there *are*
        # differences and the output should be decently readable.
        orig = test.maxDiff
        try:
            test.maxDiff = None
            test.assertMultiLineEqual(expected, actual)
        finally:
            test.maxDiff = orig


def assertLength(test, length, obj_with_len):
    """Assert that obj_with_len is of length length.

    This display 'obj_with_len' which is more helpful than just knowing that
    its expected length is wrong.
    """
    if len(obj_with_len) != length:
        test.fail('Incorrect length: wanted {}, got {}'
                  ' for {!r}'.format(length, len(obj_with_len), obj_with_len))


def assertSuccessfullTest(outer, inner):
    """Run a test and check it did so successfully.

    :param outer: The test making the assertion.

    :param inner: The test to run that should succeed.

    :return: The result object if the outer test want to do further checks.
    """
    result = unittest.TestResult()
    inner.run(result)
    outer.assertEqual(1, result.testsRun)
    outer.assertTrue(result.wasSuccessful())
    return result



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
import unittest
import subprocess
import sys
import traceback

from byot import (
    filters,
    results,
)


@contextlib.contextmanager
def catching_errors(test, result):
    """Wrap test execution catching exceptions.

    KeyboardInterrupt is re-raised, other exceptions are sent to 'result' as
    errors.

    :note: This looks like a convoluted way to call result.addError() (and it
        is) but it neatly implements the use case.
    """
    try:
        yield
    except KeyboardInterrupt:
        raise
    except:
        result.addError(test, sys.exc_info())
    return


class TestSuite(unittest.TestSuite):
    """A test suite providing setUp/addCleanup.

    ``setUp`` is called before the tests are run. If it fails, tests are not
    run. Added cleanups are always called whether or not a test or setUp
    failed.
    """

    # We fake the TestCase API (or mock the implementation details that
    # leak). It's a cheap price to be able to re-use result.addError() and all
    # good bits that comes with it.
    failureException = AssertionError

    def __init__(self, *args, **kwargs):
        super(TestSuite, self).__init__(*args, **kwargs)
        self._cleanups = []

    def id(self):
        cls = self.__class__
        # We need a unique name in case the same suite class is used several
        # times hence the id().
        return '{}.{}({:x})'.format(cls.__module__, cls.__name__, id(self))

    def run(self, result, debug=False):
        setup_done = False
        with catching_errors(self, result):
            self.setUp()
            setup_done = True
        try:
            if setup_done:
                super(TestSuite, self).run(result, debug)
        finally:
            while self._cleanups:
                cleanup, args, kwargs = self._cleanups.pop()
                with catching_errors(self, result):
                    cleanup(*args, **kwargs)

    def setUp(self):
        """Setup the test suite before running it."""
        pass

    def addCleanup(self, function, *args, **kwargs):
        """Add a cleanup function to be called after running.

        Functions added with addCleanup will be called in reverse order of
        adding after running the tests, or after setUp if setUp raises an
        exception.

        If a function added with addCleanup raises an exception, the error will
        be recorded as a test suite error, and the next cleanup will then be
        run.

        Cleanup functions are always called before a test suite finishes
        running, even if setUp is aborted by an exception.
        """
        self._cleanups.append((function, args, kwargs))


class WorkerSuite(TestSuite):
    """Suite used by workers implementing concurrent execution."""

    def __init__(self, suite, unique, factor):
        super(WorkerSuite, self).__init__()
        self.suite = suite
        self.unique = unique
        self.factor = factor
        self.stream = None


class ForkedSuite(WorkerSuite):

    def __init__(self, suite, unique, factor):
        super(ForkedSuite, self).__init__(suite, unique, factor)
        self.child_pid = None

    def run(self, result):
        read, write = os.pipe()
        self.child_pid = os.fork()
        if self.child_pid:
            os.close(write)
            self.stream = os.fdopen(read, 'rb', 0)
            try:
                self.run_parent(result)
            finally:
                os.waitpid(self.child_pid, 0)
        else:
            os.close(read)
            self.stream = os.fdopen(write, 'wb', 0)
            # Leave stderr and stdout open so we can see test noise but close
            # stdin so that the child goes away if it decides to read from
            # stdin (otherwise it's a roulette to see what child actually gets
            # keystrokes for pdb etc).

            # FIXME: This can be useful to debug if a breakpoint is set in such
            # a way that a single process will reach it. As such, it should be
            # exposed, at least so that devs can toggle it easily.
            # -- vila 2015-12-04
            sys.stdin.close()
            self.run_child(result)

    def run_parent(self, result):
        # Collect the stream sent by the child
        consumer = results.SubunitConsumer(self.stream)
        consumer.run(result)

    def run_child(self, result):
        # Send a a subunit stream to the parent, the received result is
        # ignored, it's the parent result that matters.
        try:
            result = results.SubunitProducer(self.stream)
            self.suite.run(result)
        except:
            # MISSINGTEST: Completely uncovered scenario here as we used
            # subunit v1 before -- vila 2016-01-12

            # Try and report traceback on stream, but exit with error
            # even if stream couldn't be created or something else
            # goes wrong.  The traceback is formatted to a string and
            # written in one go to avoid interleaving lines from
            # multiple failing children.
            try:
                self.stream.write(traceback.format_exc())
            finally:
                os._exit(1)
        # We're done. We don't want the execution to continue in the caller.
        os._exit(0)


class SubprocessedSuite(WorkerSuite):

    def __init__(self, suite, unique, factor):
        super(SubprocessedSuite, self).__init__(suite, unique, factor)
        self.cmd = None

    def run(self, result, debug=False):
        setup_done = False
        with catching_errors(self, result):
            self.setUp()
            setup_done = True
        try:
            if setup_done:
                process = self.run_child(result)
                # Collect the stream sent by the child
                consumer = results.SubunitConsumer(process.stdout)
                consumer.run(result)
        finally:
            while self._cleanups:
                cleanup, args, kwargs = self._cleanups.pop()
                with catching_errors(self, result):
                    cleanup(*args, **kwargs)

    def run_child(self, result):
        # FIXME: A proper error should be reported if self.cmd is not
        # set -- vila 2015-12-09
        process = subprocess.Popen(self.cmd, stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE)
        self.addCleanup(os.waitpid, process.pid, 0)
        # Close process stdin to avoid confusion when mutiple processes are
        # involved
        process.stdin.close()
        return process


class SubprocessedSuiteWithLoadFile(SubprocessedSuite):

    def __init__(self, suite, unique, factor):
        super(SubprocessedSuiteWithLoadFile, self).__init__(
            suite, unique, factor)
        self.test_list_path = None

    def setUp(self):
        super(SubprocessedSuiteWithLoadFile, self).setUp()
        self.test_list_path = '{}.test-list'.format(self.unique)
        self.addCleanup(os.remove, self.test_list_path)
        with open(self.test_list_path, 'w') as tl:
            for t in filters.iter_flat(self.suite):
                tl.write('{}\n'.format(t.id()))



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import sys
import warnings


import testtools


from byot import (
    filters,
    loaders,
    results,
    suites,
)


class PickAction(argparse.Action):
    """Implements --pick error checking and conversion.

    For a 'n/m' valid string the value will be [n, m].
    """

    def __call__(self, parser, namespace, values, option_string=None):
        value = values  # There is only one

        def BadValue(msg):
            return argparse.ArgumentError(self, msg)

        try:
            rank, modulo = value.split('/')
        except ValueError:
            raise BadValue('{} does not match n/m format'.format(value))

        def check_int(val):
            try:
                return abs(int(val))
            except ValueError:
                raise BadValue('{} is not an integer'.format(val))

        rank, modulo = [check_int(rank), check_int(modulo)]
        if not rank or not modulo:
            raise BadValue('Zero makes no sense in {}'.format(value))
        if rank > modulo:
            raise BadValue(
                '{} must be <= {} in {}'.format(rank, modulo, value))
        setattr(namespace, self.dest, [rank, modulo])


def add_byot_run_options(parser, positional=False):
    if positional:
        args = ('include_regexps',)
        kwargs = dict(nargs='*')
    else:
        args = ('--include', '-i')
        kwargs = dict(action='append', dest='include_regexps',
                      default=None)
    parser.add_argument(
        *args, metavar='INCLUDE',
        help='All tests matching the INCLUDE regexp will be run.'
        ' Can be repeated.', **kwargs)
    # Optional arguments
    parser.add_argument(
        '--module', '-m', metavar='MODULE', action='append',
        dest='modules',
        help='Load tests from MODULE[:PATH]. MODULE is searched in'
        ' python path or PATH if specified.'
        ' Can be repeated.')
    parser.add_argument(
        '--exclude', '-X', metavar='EXCLUDE', action='append',
        dest='exclude_regexps',
        help='All tests matching the EXCLUDE regexp will not be run.'
        ' Can be repeated.')
    parser.add_argument(
        '--list', '-l', action='store_true',
        dest='list_only',
        help='List the tests instead of running them.')
    parser.add_argument(
        '--load', action='store', type=argparse.FileType('r'),
        dest='load_file',
        help='A file name containing test ids to be run'
        ' in the specified order.')
    parser.add_argument(
        '--format', '-f', choices=['text', 'subunit'], default='text',
        help='Output format for the test results.')
    parser.add_argument(
        '--concurrency', '-c', dest='concurrency',
        default=1, type=int,
        help='concurrency (number of processes).')
    parser.add_argument(
        '--pick', dest='pick',
        default=None, action=PickAction,
        help='n/m pick nth test every m tests.')


class RunTestsArgParser(argparse.ArgumentParser):
    """A parser for the byot-run script.

    This can be used as a base class for scripts that want to add more options
    to suit their needs.
    """

    def __init__(self, prog=None, description=None):
        if prog is None:
            prog = 'byot-run'
        if description is None:
            description = 'Load and run tests.'
        super(RunTestsArgParser, self).__init__(prog=prog,
                                                description=description)
        add_byot_run_options(self, positional=True)

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        Overridding stdout/stderr is provided for tests.

        :params args: The arguments to the script.

        :param out: Default to sys.stdout.

        :param err: Default to sys.stderr.

        :return: The populated namespace.
        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(RunTestsArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig


def cli_run(args=None, stdout=None, stderr=None):
    """Run tests from the command line."""
    if args is None:
        args = sys.argv[1:]
    if stdout is None:
        stdout = sys.stdout
    if stderr is None:
        stderr = sys.stderr
    # FIXME: Track fix in either testtools or subunit about
    # https://bugs.launchpad.net/subunit/+bug/2045171 -- vila 2024-07-10
    warnings.filterwarnings('ignore',
                            message='TestResult has no addDuration method',
                            module='unittest.case')
    parser = RunTestsArgParser()
    ns = parser.parse_args(args)
    suite = load_tests(ns.include_regexps, ns.exclude_regexps,
                       modules=ns.modules, load_list_file=ns.load_file,
                       pick=ns.pick)
    if ns.list_only:
        ret = list_tests(suite, stdout)
    else:
        if ns.format == 'text':
            result = results.TextResult(stdout, verbosity=2)
        else:
            result = results.SubunitProducer(stdout)
        ret = run_tests(suite, result, ns.concurrency)
    return ret


def load_tests(include_regexps, exclude_regexps=None,
               modules=None, load_list_file=None, pick=None):
    """Load tests matching inclusive and exclusive regexps.

    :param include_regexps: A list of regexps describing the tests to include.

    :param exclude_regexps: A list of regexps describing the tests to exclude.

    :param modules: A list of module python names from which the tests should
        be loaded. Default to None which fallbacks to loading tests from the
        current directory.

    :param load_list_file: A file handle containing a list of test ids, one by
        line.

    :param pick: An optional (rank, modulo) tuple to pick a subset of the
        suite.

    :return: The test suite for all collected tests.
    """
    loader = loaders.Loader()
    suite = loader.suiteClass()
    if modules is None:
        here = os.path.abspath('.')
        if '' not in sys.path and here not in sys.path:
            # Loading from the current dir requires the current dir to be in
            # sys.path.
            sys.path.insert(0, here)
        suite.addTests(loader.loadTestsFromTree('.'))
    else:
        for mod_name in modules:
            mod_tests = loader.loadTestsFromSysPathModule(mod_name)
            if mod_tests is not None:
                suite.addTests(mod_tests)
    suite = filters.include_regexps(include_regexps, suite)
    suite = filters.exclude_regexps(exclude_regexps, suite)
    if load_list_file is not None:
        id_list = filters.load_test_id_list(load_list_file)
        suite = filters.include_id_list(id_list, suite)
    if pick is not None:
        rank, modulo = pick
        suite = filters.pick(suite, rank, modulo)
    return suite


def list_tests(suite, stream):
    """List the test ids , one by line.

    :param suite: A test suite to list.

    :param stream: A writable stream.

    :return: 0 on success, 1 otherwise.

    :note: Listing no tests is an error. The rationale is that when used from a
        script, most people expects to select at least one test and there has
        been numerous reports of people being confused that listing *no* tests
        wasn't flagged as an error. In most of these cases, *another* error led
        to no tests being selected but trapping it here helps.
    """
    no_tests = True
    for t in filters.iter_flat(suite):
        stream.write('{}\n'.format(t.id()))
        no_tests = False
    return int(no_tests)


def run_tests(suite, result, process_nb=1, split_suite=suites.ForkedSuite):
    """Run the provided tests with the provided test result.

    :param suite: A test suite.

    :param result: The collecting test result object.

    :param process_nb: The number of processes to split the run across.

    :param split_suite: The suite class to use for the splitted suite when
        ``nb_process`` is greater than 1.

    :return: 0 on success, 1 otherwise.

    :note: Running no tests is an error. The rationale is that when used from a
        script, most people expects to run at least one test and there has been
        numerous reports of people being confused that running *no* tests
        wasn't flagged as an error. In most of these cases, *another* error led
        to no tests being run but trapping it here helps.

    """
    if process_nb <= 1:
        result.startTestRun()
        try:
            suite.run(result)
        finally:
            result.stopTestRun()
    else:
        # Split the tests for workers
        def split_tests():
            for unique, sub_suite in enumerate(
                    filters.partition_suite(suite, process_nb), start=1):
                yield split_suite(sub_suite, unique, process_nb), unique
        # Wrap the result so it can receive a testtools stream and forward the
        # events to the received ``result``.
        conc_res = testtools.StreamToExtendedDecorator(result)
        conc_res.startTestRun()
        try:
            conc = testtools.ConcurrentStreamTestSuite(split_tests)
            conc.run(conc_res)
        finally:
            conc_res.stopTestRun()
    return int(not (result.wasSuccessful() and result.testsRun > 0))



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import unittest


import testtools


from byot import (
    assertions,
    fixtures,
    results,
)


class TestTextResultOutput(unittest.TestCase):

    def assertOutput(self, expected, kind):
        test = fixtures.make_case(kind)
        out = io.StringIO()
        res = results.TextResult(out)

        # We don't care about timing here so we always return 0 which
        # simplifies matching the expected result
        def zero(atime):
            return 0.0
        res.convert_delta_to_float = zero
        test.run(res)
        self.assertEqual(expected, res.stream.getvalue())

    def test_pass(self):
        self.assertOutput('.', 'pass')

    def test_fail(self):
        self.assertOutput('F', 'fail')

    def test_error(self):
        self.assertOutput('E', 'error')

    def test_skip(self):
        self.assertOutput('s', 'skip')

    def test_skip_reason(self):
        self.assertOutput('s', 'skip_reason')

    def test_expected_failure(self):
        self.assertOutput('x', 'expected_failure')

    def test_unexpected_success(self):
        self.assertOutput('u', 'unexpected_success')


def expand_template_for_test(template, test, kwargs=None):
    """Expand common references in template.

    Tests that check run outputs can be simplified if they use templates
    instead of litteral expected strings. There are plenty of examples below.

    :param template: A string where common strings have been replaced by a
        keyword so 1) tests are easier to read, 2) we don't run into pep8
        warnings for long lines.

    :param test: The test case under scrutiny.

    :param kwargs: A dict with more keywords for the template. This allows
        some tests to add more keywords when they are test specific.
    """
    if kwargs is None:
        kwargs = dict()

    def cleanup_file_name(name):
        # Getting the file name right is tricky, depending on whether the
        # module was just recompiled or not __file__ can be either .py or .pyc
        # but when it appears in an exception, the .py is always used.
        return name.replace('.pyc', '.py').replace('.pyo', '.py')

    filename = cleanup_file_name(__file__)
    traceback_fixed_length = kwargs.get('traceback_fixed_length', None)
    if traceback_fixed_length is not None:
        # We need to calculate the traceback detail length which includes the
        # file name.
        full_length = traceback_fixed_length + len(filename)
        # subunit prefixes the traceback with its length, this is the only
        # place that it can be properly calculated.
        kwargs['subunit_traceback_length'] = '%X' % (full_length,)
    traceback_line = getattr(test, 'traceback_line', None)
    kwargs['traceback_line'] = traceback_line
    traceback_file = getattr(test, 'traceback_file', None)
    kwargs['traceback_file'] = cleanup_file_name(traceback_file)
    # To allow easier reading for template, we format some known values
    kwargs.update(dict(classname='%s.%s' % (test.__class__.__module__,
                                            test.__class__.__name__),
                       name=test._testMethodName,
                       filename=filename))
    return template.format(**kwargs)


class TestVerboseResultOutput(unittest.TestCase):

    def assertOutput(self, template, kind):
        test = fixtures.make_case(kind)
        expected = expand_template_for_test(template, test)
        out = io.StringIO()
        res = results.TextResult(out, verbosity=2)

        # We don't care about timing here so we always return 0 which
        # simplifies matching the expected result
        def zero(atime):
            return 0.0
        res.convert_delta_to_float = zero
        test.run(res)
        self.assertEqual(expected, res.stream.getvalue())

    def test_pass(self):
        self.assertOutput('''\
{classname}.{name} ... OK (0.000 secs)
''',
                          'pass')

    def test_fail(self):
        self.assertOutput('''\
{classname}.{name} ... FAIL (0.000 secs)
''',
                          'fail')

    def test_error(self):
        self.assertOutput('''\
{classname}.{name} ... ERROR (0.000 secs)
''',
                          'error')

    def test_skip(self):
        self.assertOutput('''\
{classname}.{name} ... SKIP (0.000 secs)
''',
                          'skip')

    def test_skip_reason(self):
        self.assertOutput('''\
{classname}.{name} ... SKIP Reason (0.000 secs)
''',
                          'skip_reason')

    def test_expected_failure(self):
        self.assertOutput('''\
{classname}.{name} ... XFAIL (0.000 secs)
''',
                          'expected_failure')

    def test_unexpected_success(self):
        self.assertOutput('''\
{classname}.{name} ... NOTOK (0.000 secs)
''',
                          'unexpected_success')


def summary_from_subunit(test, stream=None, noise=None):
    """Run a suite returning the summary from the subunit v2 stream."""
    if stream is None:
        stream = io.BytesIO()
    # Create the subunit v2 stream by running the test
    res = results.SubunitProducer(stream)
    test.run(res)
    # Now, process the subunit stream so we can assert about its content
    input_stream = io.BytesIO(stream.getvalue())
    subunit_case = results.SubunitConsumer(input_stream, noise)
    summary = testtools.StreamSummary()
    summary.startTestRun()
    subunit_case.run(summary)
    summary.stopTestRun()
    return summary


class TestSubunitSummaryForUnittest(unittest.TestCase):
    """Test subunit v2 output stream."""

    def run_case(self, kind):
        case = fixtures.make_case(kind)
        summary = summary_from_subunit(case)
        self.assertEqual(1, summary.testsRun)
        return summary

    def test_pass(self):
        summary = self.run_case('pass')
        self.assertTrue(summary.wasSuccessful())

    def test_fail(self):
        summary = self.run_case('fail')
        self.assertFalse(summary.wasSuccessful())
        # FIXME: Check with newer testtools, that should be summary.failures --
        # vila 2015-05-26
        assertions.assertLength(self, 1, summary.errors)

    def test_error(self):
        summary = self.run_case('error')
        self.assertFalse(summary.wasSuccessful())
        assertions.assertLength(self, 1, summary.errors)

    def test_skip(self):
        summary = self.run_case('skip')
        self.assertTrue(summary.wasSuccessful())
        assertions.assertLength(self, 1, summary.skipped)

    def test_skip_reason(self):
        summary = self.run_case('skip_reason')
        self.assertTrue(summary.wasSuccessful())
        assertions.assertLength(self, 1, summary.skipped)
        self.assertEqual('Reason', summary.skipped[0][1])

    def test_expected_failure(self):
        summary = self.run_case('expected_failure')
        self.assertTrue(summary.wasSuccessful())
        assertions.assertLength(self, 1, summary.expectedFailures)

    def test_unexpected_success(self):
        summary = self.run_case('unexpected_success')
        self.assertTrue(summary.wasSuccessful())
        assertions.assertLength(self, 1, summary.unexpectedSuccesses)


class SpuriousOutput(unittest.TestCase):
    """What happens to the output produced by tests.

    By default, subunit/testtools will swallow the spurious output mixed with
    the subunit stream.

    SubunitConsumer explicitly routes spurious output to a dedicated stream.
    """
    def test_spurious_output(self):
        stream = io.BytesIO()
        noise = io.BytesIO()

        class Test(unittest.TestCase):

            def test_spurious(self):
                stream.write('spurious\n'.encode('utf8'))

        summary = summary_from_subunit(Test('test_spurious'), stream, noise)
        self.assertTrue(summary.wasSuccessful())
        self.assertEqual('spurious\n', noise.getvalue().decode('utf8'))

# MISSING tests: StreamResult needs to be tested somewhere. We use it
# indirectly but it could be a pain to debug -- vila 2016-01-20



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import io
import os
import tempfile
import unittest

try:
    from cStringIO import StringIO
except:
    from io import StringIO

import subunit


from byot import (
    assertions,
    features,
    fixtures,
    results,
    runners,
    tests,
)


class TestCliRunSmoke(unittest.TestCase):
    """Smoke blackbox tests."""

    def setUp(self):
        super(TestCliRunSmoke, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.out = StringIO()
        self.err = StringIO()

    def test_run_no_tests_errors(self):
        ret = runners.cli_run(['no-such-test'],
                              stdout=self.out, stderr=self.err)
        self.assertEqual(1, ret)

    def test_list_no_tests_errors(self):
        ret = runners.cli_run(['-l', 'no-such-test'],
                              stdout=self.out, stderr=self.err)
        self.assertEqual(1, ret)
        self.assertEqual('', self.out.getvalue())
        self.assertEqual('', self.err.getvalue())


class TestCliRunReal(unittest.TestCase):
    """Blackbox tests."""

    def setUp(self):
        super(TestCliRunReal, self).setUp()
        self.out = StringIO()
        self.err = StringIO()

    def test_pass(self):
        # Dummy passing test for test_list_one_test
        pass

    def test_list_one_test(self):
        # We can't use our own id or we'll loop ;)
        its_id = self.id().replace('test_list_one_test', 'test_pass')
        with tempfile.NamedTemporaryFile('w', suffix='.list',
                                         prefix='byot',
                                         delete=False) as list_file:
            self.addCleanup(os.remove, list_file.name)
            list_path = list_file.name
            list_file.write('{}\n'.format(its_id))
        ret = runners.cli_run(['--load', list_path, '--list',
                               '--module=byot'],
                              stdout=self.out, stderr=self.err)
        self.assertEqual(0, ret)
        self.assertEqual([its_id], self.out.getvalue().splitlines())
        self.assertEqual('', self.err.getvalue())


class TestListTests(unittest.TestCase):

    def setUp(self):
        super(TestListTests, self).setUp()
        fixtures.setup_for_local_imports(self)

    def list_tests(self, kinds):
        suite = fixtures.make_suite(kinds)
        out = StringIO()
        ret = runners.list_tests(suite, out)
        return ret, out.getvalue().splitlines()

    def test_no_tests_return_1(self):
        ret, test_names = self.list_tests([])
        self.assertEqual(1, ret)
        self.assertEqual([], test_names)

    def assertTestNames(self, expected, kinds):
        ret, actual = self.list_tests(kinds)
        self.assertEqual(0, ret)
        self.assertEqual(expected, actual)

    def test_single(self):
        self.assertTestNames(['byot.fixtures.UnittestTest.test_pass'],
                             ['pass'])

    def test_several(self):
        self.assertTestNames(['byot.fixtures.UnittestTest.test_pass',
                              'byot.fixtures.UnittestTest.test_fail',
                              'byot.fixtures.UnittestTest.test_skip'],
                             ['pass', 'fail', 'skip'])


class TestRunTests(unittest.TestCase):

    def setUp(self):
        super(TestRunTests, self).setUp()
        fixtures.setup_for_local_imports(self)

    def run_tests(self, kinds):
        suite = fixtures.make_suite(kinds)
        result = results.TextResult(StringIO())
        ret = runners.run_tests(suite, result)
        return ret

    def test_pass(self):
        self.assertEqual(0, self.run_tests(['pass']))

    def test_fail(self):
        self.assertEqual(1, self.run_tests(['fail']))

    def test_skip(self):
        self.assertEqual(0, self.run_tests(['skip']))


class TestRunTestsTextOutput(unittest.TestCase):

    def setUp(self):
        super(TestRunTestsTextOutput, self).setUp()
        fixtures.setup_for_local_imports(self)

    def assertOutput(self, expected, kinds, result=None):
        suite = fixtures.make_suite(kinds)
        if result is None:
            result = results.TextResult(StringIO(), verbosity=2)
        runners.run_tests(suite, result)
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              result.stream.getvalue())

    def test_pass(self):
        self.assertOutput('''Tests running...
byot.fixtures.UnittestTest.test_pass ... OK (... secs)

Ran 1 test in ...s
OK
''',
                          ['pass'])

    def test_fail(self):
        self.assertOutput('''Tests running...
byot.fixtures.UnittestTest.test_fail ... FAIL (... secs)
======================================================================
FAIL: byot.fixtures.UnittestTest.test_fail
----------------------------------------------------------------------
Traceback (most recent call last):
  File ..., in test_fail
    raise self.failureException
AssertionError

Ran 1 test in ...s
FAILED (failures=1)
''',
                          ['fail'])

    def test_skip(self):
        self.assertOutput('''Tests running...
byot.fixtures.UnittestTest.test_skip ... SKIP (... secs)

Ran 1 test in ...s
OK
1 skipped: No reason given
''',
                          ['skip'])

    def test_skip_reason(self):
        self.assertOutput('''Tests running...
byot.fixtures.UnittestTest.test_skip_reason ... SKIP Reason (... secs)

Ran 1 test in ...s
OK
1 skipped: Reason
''',
                          ['skip_reason'])

    @features.requires(tests.minimal_testtools)
    def test_fail_fast(self):
        result = results.TextResult(StringIO(), verbosity=2, failfast=True)
        self.assertOutput('''Tests running...
byot.fixtures.UnittestTest.test_fail ... FAIL (... secs)
======================================================================
FAIL: byot.fixtures.UnittestTest.test_fail
----------------------------------------------------------------------
Traceback (most recent call last):
  File ..., in test_fail
    raise self.failureException
AssertionError

Ran 1 test in ...s
FAILED (failures=1)
''',
                          ['fail', 'pass'], result=result)


class TestRunTestsSubunitOutput(unittest.TestCase):

    def setUp(self):
        super(TestRunTestsSubunitOutput, self).setUp()
        fixtures.setup_for_local_imports(self)

    def assertOutput(self, expected, kinds, result=None):
        suite = fixtures.make_suite(kinds)
        if result is None:
            if tests.minimal_testtools.available():
                stream = io.BytesIO()
            else:
                stream = StringIO()
            result = subunit.TestProtocolClient(stream)
        runners.run_tests(suite, result)
        actual = result._stream.getvalue().decode('utf8')
        assertions.assertMultiLineAlmostEqual(self, expected, actual)

    @features.requires(tests.minimal_testtools)
    def test_pass(self):
        self.assertOutput('''test: byot.fixtures.UnittestTest.test_pass
successful: byot.fixtures.UnittestTest.test_pass
''',
                          ['pass'])

    @features.requires(tests.minimal_testtools)
    def test_fail(self):
        self.assertOutput('''test: byot.fixtures.UnittestTest.test_fail
failure: byot.fixtures.UnittestTest.test_fail [
Traceback (most recent call last):
  File ..., in test_fail
    raise self.failureException
AssertionError
]
''',
                          ['fail'])

    @features.requires(tests.minimal_testtools)
    def test_skip(self):
        self.assertOutput('''test: byot.fixtures.UnittestTest.test_skip
skip: byot.fixtures.UnittestTest.test_skip [

]
''',
                          ['skip'])

    @features.requires(tests.minimal_testtools)
    def test_fail_fast(self):
        result = subunit.TestProtocolClient(io.BytesIO())
        result.failfast = True
        self.assertOutput('''test: byot.fixtures.UnittestTest.test_fail
failure: byot.fixtures.UnittestTest.test_fail [
Traceback (most recent call last):
  File ..., in test_fail
    raise self.failureException
AssertionError
]
''',
                          ['fail', 'pass'], result=result)


class TestOptionParsing(unittest.TestCase):

    def setUp(self):
        super(TestOptionParsing, self).setUp()
        self.out = StringIO()
        self.err = StringIO()

    def parse_args(self, args):
        ns = runners.RunTestsArgParser().parse_args(args, self.out, self.err)
        return ns

    def test_help(self):
        fixtures.isolate_from_env(self)
        # Make sure the console width stays the same
        fixtures.override_env(self, 'COLUMNS', '80')
        with self.assertRaises(SystemExit):
            self.parse_args(['-h'])
        self.maxDiff = None
        assertions.assertMultiLineAlmostEqual(self, '''\
usage: byot-run [-h] [--module MODULE] [--exclude EXCLUDE] [--list]
                [--load LOAD_FILE] [--format {format}]
                [--concurrency CONCURRENCY] [--pick PICK]
...

Load and run tests.

positional arguments:
  INCLUDE               All tests matching the INCLUDE regexp will be run. Can
                        be repeated.

{argparse_options}:
  -h, --help            show this help message and exit
  --module, -m MODULE   Load tests from MODULE[:PATH]. MODULE is searched in
                        python path or PATH if specified. Can be repeated.
  --exclude, -X EXCLUDE
                        All tests matching the EXCLUDE regexp will not be run.
                        Can be repeated.
  --list, -l            List the tests instead of running them.
  --load LOAD_FILE      A file name containing test ids to be run in the
                        specified order.
  --format, -f {format}
                        Output format for the test results.
  --concurrency, -c CONCURRENCY
                        concurrency (number of processes).
  --pick PICK           n/m pick nth test every m tests.
'''.format(argparse_options=tests.python.argparse_options,
           format='{text,subunit}'), self.out.getvalue())

    def test_default_values(self):
        ns = self.parse_args([])
        self.assertEqual([], ns.include_regexps)
        self.assertEqual(None, ns.modules)
        self.assertEqual(None, ns.exclude_regexps)
        self.assertFalse(ns.list_only)
        self.assertEqual('text', ns.format)
        self.assertEqual(1, ns.concurrency)
        self.assertEqual(None, ns.load_file)
        self.assertEqual(None, ns.pick)

    def test_modules(self):
        ns = self.parse_args(['--module', 'a', '-m', 'b:.'])
        self.assertEqual(['a', 'b:.'], ns.modules)

    def test_include_regexps(self):
        ns = self.parse_args(['a', 'b'])
        self.assertEqual(['a', 'b'], ns.include_regexps)

    def test_exclude_regexps(self):
        ns = self.parse_args(['--exclude', 'a', '-X', 'b'])
        self.assertEqual(['a', 'b'], ns.exclude_regexps)

    def test_list(self):
        ns = self.parse_args(['--list'])
        self.assertTrue(ns.list_only)
        ns = self.parse_args(['-l'])
        self.assertTrue(ns.list_only)

    def test_format_text(self):
        ns = self.parse_args(['--format', 'text'])
        self.assertEqual('text', ns.format)

    def test_format_subunit(self):
        ns = self.parse_args(['-f', 'subunit'])
        self.assertEqual('subunit', ns.format)

    def test_load_list_existing(self):
        fixtures.set_uniq_cwd(self)
        with io.open('list', 'w', encoding='utf-8') as f:
            f.write('\n')
        ns = self.parse_args(['--load', 'list'])
        self.assertEqual('list', ns.load_file.name)

    def test_load_list_unknown(self):
        fixtures.set_uniq_cwd(self)
        with self.assertRaises(SystemExit):
            self.parse_args(['--load', 'I-dont-exist'])
        last_line = self.err.getvalue().splitlines()[-1]
        self.assertTrue(last_line.endswith("'I-dont-exist'"))

    def test_valid_pick(self):
        ns = self.parse_args(['--pick', '1/2'])
        self.assertEqual([1, 2], ns.pick)

    def test_pick_absolute(self):
        ns = self.parse_args(['--pick', '1/-2'])
        self.assertEqual([1, 2], ns.pick)

    def test_pick_no_slash(self):
        with self.assertRaises(SystemExit):
            self.parse_args(['--pick', '12'])

    def test_pick_requires_no_zero(self):
        with self.assertRaises(SystemExit):
            self.parse_args(['--pick', '0/1'])
        with self.assertRaises(SystemExit):
            self.parse_args(['--pick', '1/0'])

    def test_pick_rank_too_big(self):
        with self.assertRaises(SystemExit):
            self.parse_args(['--pick', '2/1'])


class TestArgParserSubClassing(unittest.TestCase):

    def test_new_prog(self):

        class New(runners.RunTestsArgParser):

            def __init__(self, prog):
                super(New, self).__init__(prog)

        new = New('foo')
        self.assertEqual('foo', new.prog)

    def test_new_description(self):
        class New(runners.RunTestsArgParser):

            def __init__(self, description):
                super(New, self).__init__(description=description)

        new = New('foo')
        self.assertEqual('foo', new.description)
        # 'prog' is unchanged
        self.assertEqual('byot-run', new.prog)


class TestAddByotOptions(unittest.TestCase):

    def setUp(self):
        super(TestAddByotOptions, self).setUp()

        class New(runners.RunTestsArgParser):

            def __init__(self):
                # Shorcut runners.RunTestsArgParser __init__ so we can call
                # add_byot_tests_options as needed
                argparse.ArgumentParser.__init__(
                    self, prog='foo', description='bar')
                runners.add_byot_run_options(self)
        self.parser = New()
        self.out = StringIO()
        self.err = StringIO()

    def parse_args(self, args):
        ns = self.parser.parse_args(args, self.out, self.err)
        return ns

    def test_empty(self):
        ns = self.parse_args([])
        self.assertEqual(None, ns.include_regexps)

    def test_include(self):
        my_id = self.id()
        ns = self.parse_args(['--include', my_id, '-i', 'foo'])
        self.assertEqual([my_id, 'foo'], ns.include_regexps)



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from byot import assertions


class TestAssertMultiLineAlmostEqual(unittest.TestCase):

    def check(self, expected, actual):
        assertions.assertMultiLineAlmostEqual(self, expected, actual)

    def test_empty(self):
        self.check('', '')

    def test_single_ellipsis_single_line(self):
        self.check('...', 'tagada')

    def test_single_ellipsis_multiple_lines(self):
        self.check('...', 'foo\nbar')

    def test_multiple_ellipsis_multiple_lines(self):
        self.check('...\n...', 'foo\nbar')

    def test_traceback(self):
        # One main target of the assertion is to be able to match tracebacks
        # without having to care about file names/lines.
        self.check('''Traceback (most recent call last):
  File ..., in expectFailure
    predicate(*args, **kwargs)
  File ..., in assertEqual
    self.assertThat(observed, matcher, message)
  File ..., in assertThat
    raise MismatchError(matchee, matcher, mismatch, verbose)
testtools.matchers._impl.MismatchError: 1 != 0
''',
                   # This is almost a real traceback but the file names have
                   # been shortened to silent pep8 warnings about long lines ;)
                   '''Traceback (most recent call last):
  File "testtools/testcase.py", line 441, in expectFailure
    predicate(*args, **kwargs)
  File "testtools/testcase.py", line 322, in assertEqual
    self.assertThat(observed, matcher, message)
  File "testtools/testcase.py", line 417, in assertThat
    raise MismatchError(matchee, matcher, mismatch, verbose)
testtools.matchers._impl.MismatchError: 1 != 0
''')


class TestAssertLength(unittest.TestCase):

    def test_assertLength_matches_empty(self):
        assertions.assertLength(self, 0, [])

    def test_assertLength_matches_nonempty(self):
        assertions.assertLength(self, 3, [1, 2, 3])

    def test_assertLength_fails_different(self):
        self.assertRaises(AssertionError, assertions.assertLength, self, 1, [])

    def test_assertLength_shows_sequence_in_failure(self):
        with self.assertRaises(AssertionError) as cm:
            assertions.assertLength(self, 2, [1, 2, 3])
        self.assertEqual('Incorrect length: wanted 2, got 3 for [1, 2, 3]',
                         cm.exception.args[0])



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import re


class NameMatcher(object):
    """Defines rules to select names.

    The rules are defined by two lists of regular expressions:
    - includes: matching succeeds if one of the regular expression match.
    - excludes: matching fails if one of the regular expression match.

    Matching fails if no rules are given.
    """

    def __init__(self, includes=None, excludes=None):
        self.includes = []
        if includes is not None:
            for inc in includes:
                self.includes.append(re.compile(inc))
        self.excludes = []
        if excludes is not None:
            for exc in excludes:
                self.excludes.append(re.compile(exc))

    def matches(self, name):
        for exc in self.excludes:
            if exc.search(name):
                return False
        for inc in self.includes:
            if inc.search(name):
                return True
        # Not explicitely included
        return False



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013, 2014, 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import io
import os
import unittest


import pep8


from byot import (
    assertions,
    features,
    matchers,
    tests,
    walkers,
)


if tests.minimal_pep8.available():
    BaseReport = pep8.BaseReport
else:
    BaseReport = object


class UnitTestReport(BaseReport):

    def __init__(self, options):
        super(UnitTestReport, self).__init__(options)
        self._fmt = pep8.REPORT_FORMAT.get(options.format.lower(),
                                           options.format)
        self._msgs = []

    def error(self, line_number, offset, text, check):
        """Report an error, according to options."""
        super(UnitTestReport, self).error(line_number, offset, text, check)
        self._msgs.append(self._fmt % {
            'path': self.filename,
            'row': self.line_offset + line_number, 'col': offset + 1,
            'code': text[:4], 'text': text[5:],
        })


@features.requires(tests.minimal_pep8)
class TestPep8(unittest.TestCase):
    """Base test class to check a list of python modules with pep8.

    :cvar packages: A list of modules defined by the daughter class. All python
        files below this module will be checked.
    """

    packages = []
    exclude = []

    def setUp(self):
        super(TestPep8, self).setUp()
        self.pep8style = pep8.StyleGuide(
            exclude=self.exclude,
            filename=['*.py'],
            show_source=True,
        )
        self.report = self.pep8style.init_report(UnitTestReport)

    def test_pep8_conformance(self):
        self.assertNotEqual([], self.packages,
                            'You should define some packages to check')
        for package in self.packages:
            self.pep8style.input_dir(os.path.dirname(package.__file__))
        self.assertEqual([], self.report._msgs,
                         '\n'.join(self.report._msgs))


@features.requires(tests.minimal_pyflakes)
class TestPyflakes(unittest.TestCase):
    """Base test class to check a list of python modules with pyflakes.

    :cvar packages: A list of modules defined by the daughter class. All python
        files below this module will be checked.

    :cvar excludes: A list of paths that should not be checked.
    """

    packages = []
    excludes = []

    def test_pyflakes_conformance(self):
        from pyflakes import (
            api,
            reporter,
        )
        self.assertNotEqual([], self.packages,
                            'You should define some packages to check')
        out = io.StringIO()
        err = io.StringIO()

        # Workaround a bug in pyflakes on vivid -- vila 2015-08-21
        # still there on wily -- vila 2015-12-04
        class Reporter(reporter.Reporter):

            def flake(self, message):
                self._stdout.write('{}'.format(message))
                self._stdout.write('\n')

        report = Reporter(out, err)
        paths = [os.path.dirname(p.__file__) for p in self.packages]

        walker = PythonFileWalker(None)
        for p in paths:
            root, base = os.path.split(p)
            sw = walker.SubWalker(root)
            pyflakes_check_dir(sw, base, self.excludes,
                               api.checkPath, report)
        out_val = out.getvalue()
        if out_val:
            assertions.assertMultiLineAlmostEqual(self, '', out_val)
        err_val = err.getvalue()
        if err_val:
            assertions.assertMultiLineAlmostEqual(self, '', err_val)


def PythonFileWalker(root):
    file_matcher = matchers.NameMatcher(includes=[r'^.*\.py$'])

    def rel_file_name(walker, dir_path, name):
        return os.path.join(dir_path, name)

    def recurse_for_files(walker, dir_path):
        # Recurse and yield all items below dir_path (dir_path itself is
        # not yielded)
        for item in walker.iter(dir_path):
            yield item

    walker = walkers.Walker(root, file_matcher=file_matcher,
                            sort_key=str,
                            file_handler=rel_file_name,
                            dir_handler=recurse_for_files)
    return walker


def pyflakes_check_dir(walker, dir_path, excludes, check, report):
    for file_path in walker.iter(dir_path):
        if file_path not in excludes:
            check(walker.join_root(file_path), report)



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013, 2014, 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import sys

from byot import (
    assertions,
    filters,
    fixtures,
    loaders,
    matchers,
    tests,
)


class TestImportFromPath(unittest.TestCase):

    def setUp(self):
        super(TestImportFromPath, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()

    def test_invalid_path(self):
        with open('__init__.py', 'w') as f:
            f.write('')
        with self.assertRaises(TypeError) as cm:
            self.loader.importFromPath('.')
        self.assertEqual(
            "the 'package' argument is required to perform a relative "
            "import for '.'", str(cm.exception))

    def test_weird_path(self):
        d = './foo.bar/baz'
        os.makedirs(d)
        with open(os.path.join(d, '__init__.py'), 'w') as f:
            f.write('')
        with self.assertRaises(ImportError) as cm:
            self.loader.importFromPath(d)
        assertions.assertMultiLineAlmostEqual(self, '''\
Failed to import foo.bar.baz at ./foo.bar/baz:
Traceback (most recent call last):
  File "...", line ..., in importFromPath
    return importlib.import_module(mod_name)
{backtrace_carets}  File "...", line ..., in import_module
    return ...
{backtrace_carets}  ...
  ...
  ...
  ...
  ...
  ...
  ...
  ...
  ...
  ...
  ...
{mnf_error}: No module named 'foo'
'''.format(**tests.python.__dict__), str(cm.exception))

    def test_invalid_python_file(self):
        with open('foo.py', 'w') as f:
            f.write("I'm no python code")
        with self.assertRaises(SyntaxError):
            self.loader.importFromPath('./foo.py')

    def test_invalid_file(self):
        with open('foo', 'w') as f:
            f.write("I'm no python code and I'm not even pretending")
        with self.assertRaises(ImportError) as cm:
            self.loader.importFromPath('./foo')
        assertions.assertMultiLineAlmostEqual(self, '''\
Failed to import foo at ./foo:
Traceback (most recent call last):
  File "...", line ..., in importFromPath
    return importlib.import_module(mod_name)
{backtrace_carets}  File "...", line ..., in import_module
  ...
  ...
  ...
  ...
{mnf_error}: No module named 'foo'
'''.format(**tests.python.__dict__), str(cm.exception))

    def test_unknown_file(self):
        with self.assertRaises(ImportError) as cm:
            self.loader.importFromPath('./foo.py')
        assertions.assertMultiLineAlmostEqual(self, '''\
Failed to import foo at ./foo.py:
Traceback (most recent call last):
  File "...", line ..., in importFromPath
    return importlib.import_module(mod_name)
{backtrace_carets}  File "...", line ..., in import_module
  ...
  ...
  ...
  ...
{mnf_error}: No module named 'foo'
'''.format(**tests.python.__dict__), str(cm.exception))

    def test_valid_file(self):
        self.assertIs(None, sys.modules.get('foo', None))
        with open('foo.py', 'w') as f:
            f.write("a = 'bar'")
        self.loader.importFromPath('./foo.py')
        # We can't directly use 'foo.a' but the module has been imported
        self.assertEqual('bar', sys.modules['foo'].a)


class TestPackagePathFromName(unittest.TestCase):

    def setUp(self):
        super(TestPackagePathFromName, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader(root=self.uniq_dir)

    def test_simple_package_name(self):
        fixtures.build_tree('''\
dir: foo
file: foo/__init__.py
imported = True
''')
        sys_path_entry, rel_path = self.loader.packageSysPathFromName('foo')
        self.assertEqual(self.uniq_dir, sys_path_entry)
        self.assertEqual('foo', rel_path)
        self.assertTrue(sys.modules['foo'].imported)

    def test_dotted_package_name(self):
        fixtures.build_tree('''\
dir: foo
file: foo/__init__.py
dir: foo/bar
file: foo/bar/__init__.py
imported = True
''')
        sys_path_entry, rel_path = self.loader.packageSysPathFromName(
            'foo.bar')
        self.assertEqual(self.uniq_dir, sys_path_entry)
        self.assertEqual('foo/bar', rel_path)
        self.assertTrue(sys.modules['foo.bar'].imported)

    def test_failing_import(self):
        with self.assertRaises(ImportError) as cm:
            sys_path_entry, rel_path = self.loader.packageSysPathFromName(
                'foo')
        self.assertEqual("No module named 'foo'", str(cm.exception))


class TestSortNames(unittest.TestCase):
    """Smoke test the sortNames method."""

    def setUp(self):
        super(TestSortNames, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()

    def test_sort_empty(self):
        self.assertEqual([], self.loader.sortNames([]))

    def test_sort_names(self):
        self.assertEqual(['a', 'b'], self.loader.sortNames(['b', 'a']))


def assertTestsEqual(test, expected, actual):
    """Assert that a test [suite] is the expected one.

    :param expected: A flat list of tests.

    :param actual: A TestSuite. Since this is flattened, it can also be a
        single test or a list of tests.
    """
    test.assertEqual(list(filters.iter_flat(expected)),
                     list(filters.iter_flat(actual)))


def assertTestIDs(test, expected, actual):
    """Assert that the test ids are the expected ones.

    :param expected: A list of test ids.

    :param actual: A TestSuite. Since this is flattened, it can also be a
        single test or a list of tests.
    """
    test.assertEqual(expected, [t.id()
                                for t in list(filters.iter_flat(actual))])


class TestIterFlat(unittest.TestCase):
    """Test iter_flat through assertTestsEqual acting as a proxy.

    This keep the tests short while giving 'assertTestsEqual' examples.
    """

    def test_empty_lists(self):
        assertTestsEqual(self, [], [])

    def test_empty_suite(self):
        assertTestsEqual(self, [], unittest.TestSuite())

    def test_single_vs_single(self):
        test = fixtures.make_case()
        assertTestsEqual(self, test, test)

    def test_single_vs_list(self):
        test = fixtures.make_case()
        assertTestsEqual(self, test, [test])

    def test_list_vs_single(self):
        test = fixtures.make_case()
        assertTestsEqual(self, [test], test)

    def test_list_vs_list(self):
        test = fixtures.make_case()
        assertTestsEqual(self, [test], [test])

    def test_simple_suite(self):
        test = fixtures.make_case()
        assertTestsEqual(self, test, unittest.TestSuite([test]))

    def test_embedded_suite(self):
        test = fixtures.make_case()
        inner = unittest.TestSuite([test])
        middle = unittest.TestSuite([inner])
        assertTestsEqual(self, test, unittest.TestSuite([middle]))


class TestAssertTestIDs(unittest.TestCase):

    def test_single(self):
        assertTestIDs(self, ['byot.fixtures.UnittestTest.test_pass'],
                      fixtures.make_case())

    def test_several(self):
        assertTestIDs(self, ['byot.fixtures.UnittestTest.test_pass',
                             'byot.fixtures.UnittestTest.test_pass'],
                      [fixtures.make_case(), fixtures.make_case()])


class TestLoadTestsFromFile(unittest.TestCase):

    def setUp(self):
        super(TestLoadTestsFromFile, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()

    def test_empty_file(self):
        pass

    def test_single_test(self):
        fixtures.build_tree('''file: test_foo.py
import unittest

class Test(unittest.TestCase):

    def test_me(self):
      self.assertTrue(True)
''')
        suite = self.loader.loadTestsFromFile('./test_foo.py')
        self.assertEqual(1, suite.countTestCases())

    def test_sorted_tests(self):
        fixtures.build_tree('''file: test_foo.py
import unittest

class TestB(unittest.TestCase):

    def test_me_1(self):
      self.assertTrue(True)

class TestC(TestB):

    def test_me_2(self):
      self.assertTrue(True)

class TestA(TestC):

    def test_me_3(self):
        self.assertFalse(False)
''')
        suite = self.loader.loadTestsFromFile('./test_foo.py')
        self.assertEqual(6, suite.countTestCases())
        self.assertEqual(['test_foo.TestA.test_me_1',
                          'test_foo.TestA.test_me_2',
                          'test_foo.TestA.test_me_3',
                          'test_foo.TestB.test_me_1',
                          'test_foo.TestC.test_me_1',
                          'test_foo.TestC.test_me_2'],
                         [t.id() for t in filters.iter_flat(suite)])


class TestLoadTestFromFiles(unittest.TestCase):

    def setUp(self):
        super(TestLoadTestFromFiles, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()

    def test_empty_list(self):
        fixtures.build_tree('''file: p.py
import unittest

class Test(unittest.TestCase):

    def test_pass(self):
        pass
''')
        assertTestsEqual(self, [], self.loader.loadTestsFromFiles('.', []))

    def test_single_file(self):
        fixtures.build_tree('''file: test.py
import unittest

class Test(unittest.TestCase):

    def test_pass(self):
        pass
''')
        assertTestIDs(self, ['test.Test.test_pass'],
                      self.loader.loadTestsFromFiles('.', ['test.py']))

    def test_several_files(self):
        fixtures.build_tree('''
file: test2.py
import unittest

class Test(unittest.TestCase):

    def test_pass(self):
        pass
file: test1.py
import unittest

class Test(unittest.TestCase):

    def test_pass(self):
        pass
''')
        assertTestIDs(
            self, ['test1.Test.test_pass', 'test2.Test.test_pass'],
            self.loader.loadTestsFromFiles('.', ['test1.py', 'test2.py']))

    def test_empty_dir(self):
        assertTestsEqual(self, [], self.loader.loadTestsFromFiles('.', []))


class TestLoadTestFromFiles_Matching(unittest.TestCase):

    def setUp(self):
        super(TestLoadTestFromFiles_Matching, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()
        # A simple tree fitting all tests. Pro: single context to think about,
        # smaller tests. Cons: shared resource, do not abuse. Specific test are
        # encouraged to enrich locally to suit their needs rather than pollute
        # the shared env.
        fixtures.build_tree('''
file: other.py
import unittest
class T(unittest.TestCase):
  def test_inc(self): pass
file: inc.py
import unittest
class T(unittest.TestCase):
  def test_inc(self): pass
file: exc.py
import unittest
class T(unittest.TestCase):
  def test_exc(self): pass
dir: other_dir
file: other_dir/__init__.py
file: other_dir/p.py
import unittest
class T(unittest.TestCase):
  def test_inc(self): pass
dir: inc_dir
file: inc_dir/__init__.py
file: inc_dir/p.py
import unittest
class T(unittest.TestCase):
  def test_inc(self): pass
dir: exc_dir
file: exc_dir/__init__.py
file: exc_dir/p.py
import unittest
class T(unittest.TestCase):
  def test_exc(self): pass
''')

    def assertIDs(self, expected, file_matching=None, dir_matching=None):
        # Matchers collect all by default so test can precisely control what
        # they get. This differs from the Loader class default which provides a
        # better user experience.
        if file_matching is None:
            file_matching = dict(includes=[r'.*\.py$'])
        if dir_matching is None:
            dir_matching = dict(includes=['.*'])
        loader = self.loader.SubLoader(
            dir_matcher=matchers.NameMatcher(**dir_matching),
            file_matcher=matchers.NameMatcher(**file_matching))
        actual = loader.loadTestsFromFiles('.', os.listdir('.'))
        assertTestIDs(self, expected, actual)

    def test_not_maching_file(self):
        self.assertIDs([], dict(includes=['doesnt_exist']))

    def test_included_file(self):
        self.assertIDs(['inc.T.test_inc'],
                       file_matching=dict(includes=['inc.py']))

    def test_excluded_file(self):
        self.assertIDs(['exc_dir.p.T.test_exc',
                        'inc.T.test_inc', 'inc_dir.p.T.test_inc',
                        'other.T.test_inc', 'other_dir.p.T.test_inc'],
                       # Since matching fails if no rules are given, excluding
                       # one file requires including all files first.
                       file_matching=dict(excludes=['exc.py'],
                                          includes=[r'.*\.py$']))

    def test_not_matching_dir(self):
        # We get only the tests in the files at the top level
        self.assertIDs(['exc.T.test_exc', 'inc.T.test_inc',
                        'other.T.test_inc'],
                       dir_matching=dict(includes=['doesntexist']))

    def test_included_dir(self):
        self.assertIDs(['exc.T.test_exc',
                        'inc.T.test_inc', 'inc_dir.p.T.test_inc',
                        'other.T.test_inc'],
                       dir_matching=dict(includes=['inc_dir']))

    def test_excluded_dir(self):
        self.assertIDs(['exc.T.test_exc',
                        'inc.T.test_inc', 'inc_dir.p.T.test_inc',
                        'other.T.test_inc', 'other_dir.p.T.test_inc'],
                       # Since matching fails if no rules are given, excluding
                       # one dir requires including all dirs first.
                       dir_matching=dict(excludes=['exc_dir'],
                                         includes=['.*']))

    def test_all(self):
        # No matching specified, we get all tests
        self.assertIDs(['exc.T.test_exc', 'exc_dir.p.T.test_exc',
                        'inc.T.test_inc', 'inc_dir.p.T.test_inc',
                        'other.T.test_inc', 'other_dir.p.T.test_inc'])


class TestLoadTestsFromPackage(unittest.TestCase):

    def setUp(self):
        super(TestLoadTestsFromPackage, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()

    def test_no_package(self):
        self.assertEqual(None, self.loader.loadTestsFromPackage('.'))

    def test_no_init_file(self):
        fixtures.build_tree('''
dir: t
''')
        self.assertEqual(None, self.loader.loadTestsFromPackage('t'))

    def test_import_failure(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
import not_a_python_package
''')
        with self.assertRaises(ImportError):
            self.loader.loadTestsFromPackage('t')

    def test_syntax_error(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
Do I look like python code ?
''')
        with self.assertRaises(SyntaxError):
            self.loader.loadTestsFromPackage('t')

    def test_empty_package(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py''')
        assertTestIDs(self, [], self.loader.loadTestsFromPackage('t'))

    def test_init_tests(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
import unittest
class T(unittest.TestCase):
  def test_init(self): pass
''')
        assertTestIDs(self, ['t.T.test_init'],
                      self.loader.loadTestsFromPackage('t'))

    def test_file_tests(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
file: t/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        assertTestIDs(self, ['t.test.T.test_pass'],
                      self.loader.loadTestsFromPackage('t'))

    def test_only_init_tests(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
def load_tests(loader, std_tests, pattern):
    # We only return the tests defined in __init__.py ignoring anything else
    return std_tests
import unittest
class T(unittest.TestCase):
  def test_t(self): pass
file: t/ignored.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        # Since load_tests is reponsible for all the tests in the tree, the
        # other files are ignored.
        assertTestIDs(self, ['t.T.test_t'],
                      self.loader.loadTestsFromPackage('t'))

    def test_init_and_file_tests(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
import unittest
class T(unittest.TestCase):
  def test_t(self): pass
file: t/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        assertTestIDs(self, ['t.T.test_t', 't.test.T.test_pass'],
                      self.loader.loadTestsFromPackage('t'))

    def test_none_from_load_tests(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
def load_tests(loader, std_tests, pattern):
    return None
import unittest
class T(unittest.TestCase):
  def test_t(self): pass
file: t/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        # Returning None is obeyed. No other loading is attempted
        self.assertEqual(None, self.loader.loadTestsFromPackage('t'))

    def test_load_tests_overrides(self):
        fixtures.build_tree('''
dir: t
file: t/__init__.py
def load_tests(loader, std_tests, pattern):
    return loader.loadTestsFromFiles('t', ['test.py'])
import unittest
class T(unittest.TestCase):
  def test_t(self): pass
file: t/ignored.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
file: t/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        assertTestIDs(self, ['t.test.T.test_pass'],
                      self.loader.loadTestsFromPackage('t'))


class TestLoadTestsFromTree(unittest.TestCase):

    def setUp(self):
        super(TestLoadTestsFromTree, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()
        # A common base for tests to build upon
        fixtures.build_tree('''
dir: t
file: t/__init__.py
file: t/ignored.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
file: t/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')

    def test_empty(self):
        fixtures.build_tree('''dir: empty''')
        assertTestIDs(self, [], self.loader.loadTestsFromTree('empty'))

    def test_none_from_load_tests(self):
        fixtures.build_tree('''
file: t/__init__.py
def load_tests(loader, std_tests, pattern):
    return None
import unittest
class T(unittest.TestCase):
  def test_t(self): pass
''')
        # Returning None is the same as having an empty module: the loading
        # falls back to iterating the other files.
        assertTestIDs(self, ['t.test.T.test_pass'],
                      self.loader.loadTestsFromTree('.'))

    def test_load_tests_explicit_file(self):
        fixtures.build_tree('''
file: t/__init__.py
def load_tests(loader, std_tests, pattern):
    return loader.loadTestsFromFiles('t', ['test.py'])
import unittest
class T(unittest.TestCase):
  def test_t(self): pass
''')
        assertTestIDs(self, ['t.test.T.test_pass'],
                      self.loader.loadTestsFromTree('.'))

    def test_load_tests_explicit_dir(self):
        fixtures.build_tree('''
file: t/__init__.py
def load_tests(loader, std_tests, pattern):
    return loader.loadTestsFromFiles('t/sub', ['test.py'])
dir: t/sub
file: t/sub/__init__.py
import unittest
class T(unittest.TestCase):
  def test_t(self): pass
file: t/sub/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        # We get the test in 't.sub.test' and ignore the others
        assertTestIDs(self, ['t.sub.test.T.test_pass'],
                      self.loader.loadTestsFromTree('.'))


class TestSubLoader(unittest.TestCase):

    def setUp(self):
        super(TestSubLoader, self).setUp()
        fixtures.setup_for_local_imports(self)
        self.loader = loaders.Loader()
        # A common base for tests to build upon
        fixtures.build_tree('''
dir: t
file: t/__init__.py
file: t/ignored.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
file: t/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')

    def test_load_tests_new_matching_files(self):
        fixtures.build_tree('''
file: t/__init__.py
import os
from byot import loaders, matchers
def load_tests(loader, std_tests, pattern):
    subloader = loader.SubLoader(
        file_matcher=matchers.NameMatcher(includes=[r'^t\.py$']))
    return subloader.loadTestsFromFiles('t', os.listdir('t'))
file: t/t.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        # We get the test in 't/t.py' and ignore the others
        assertTestIDs(self, ['t.t.T.test_pass'],
                      self.loader.loadTestsFromTree('.'))

    def test_load_tests_new_matching_dirs(self):
        fixtures.build_tree('''
file: t/__init__.py
import os
from byot import loaders, matchers
def load_tests(loader, std_tests, pattern):
    subloader = loader.SubLoader(
        dir_matcher=matchers.NameMatcher(includes=[r'^t1$']))
    return subloader.loadTestsFromFiles('t', os.listdir('t'))
dir: t/t1
file: t/t1/__init__.py
file: t/t1/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
dir: t/t2
file: t/t2/__init__.py
file: t/t2/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        # We get the test in 't/test.py' because it's loaded before the dir
        # matching applies, 't/t1/test.py' and ignore the others
        assertTestIDs(self, ['t.t1.test.T.test_pass', 't.test.T.test_pass'],
                      self.loader.loadTestsFromTree('.'))

    def test_load_single_test_dir_in_a_root(self):
        fixtures.build_tree('''
dir: t/t1
file: t/t1/__init__.py
file: t/t1/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
dir: t/t2
file: t/t2/__init__.py
file: t/t2/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        root = os.path.join(os.getcwd(), 't')
        sys.path.insert(0, root)
        subloader = self.loader.SubLoader(root=root)
        assertTestIDs(self, ['t1.test.T.test_pass'],
                      subloader.loadTestsFromPackage('t1'))

    def test_load_from_sys_path_module(self):
        fixtures.build_tree('''
dir: t/t1
file: t/t1/__init__.py
file: t/t1/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
dir: t/t2
file: t/t2/__init__.py
file: t/t2/test.py
import unittest
class T(unittest.TestCase):
  def test_pass(self): pass
''')
        sys.path.pop(0)  # Remove the current directory
        sys.path.insert(0, os.path.join(os.getcwd(), 't'))
        assertTestIDs(self, ['t1.test.T.test_pass'],
                      self.loader.loadTestsFromSysPathModule('t1'))
        assertTestIDs(self, ['t2.test.T.test_pass'],
                      self.loader.loadTestsFromSysPathModule('t2'))



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import re
import unittest


from byot import (
    assertions,
    filters,
    fixtures,
)


def create_tests_from_ids(ids):
    """Create TestCase objects from a list of ids (strings)."""
    suite = unittest.TestSuite()

    def test_id(name):
        return lambda: name

    for tid in ids:
        # We need an existing method to create a test. Arbitrarily, we use
        # id(), that souldn't fail ;) We won't run the test anyway.
        test = unittest.TestCase(methodName='id')
        # We can't define the lambda here or 'name' stay bound to the
        # variable instead of the value, use a proxy to capture the value.
        test.id = test_id(tid)
        suite.addTest(test)
    return suite


class TestFilterTestsById(unittest.TestCase):

    def assertFiltered(self, expected, condition, ids):
        """Check that ``condition`` filters tests created from ``ids``."""
        filtered = filters.filter_suite(condition, create_tests_from_ids(ids))
        self.assertEqual(expected, [t.id() for t in filtered])

    def test_filter_none(self):
        test_names = ['foo', 'bar']
        self.assertFiltered(test_names, lambda t: True, test_names)

    def test_filter_all(self):
        test_names = ['foo', 'bar']
        self.assertFiltered([], lambda t: False, test_names)

    def test_filter_start(self):
        self.assertFiltered(['foo', 'footix'],
                            lambda t: t.id().startswith('foo'),
                            ['foo', 'footix', 'bar', 'baz', 'fo'])

    def test_filter_in(self):
        self.assertFiltered(['bar', 'baz'],
                            lambda t: t.id() in ('bar', 'baz'),
                            ['foo', 'footix', 'bar', 'baz', 'fo'])

    def test_filter_single(self):
        self.assertFiltered(['bar'],
                            lambda t: t.id() == 'bar',
                            ['foo', 'bar', 'baz'])

    def test_filter_regexp(self):
        ba = re.compile('ba')
        self.assertFiltered(['bar', 'baz', 'foobar'],
                            lambda t: bool(ba.search(t.id())),
                            ['foo', 'bar', 'baz', 'foobar', 'qux'])


class TestFilterTestsByRegexps(unittest.TestCase):

    def assertFiltered(self, expected, regexps, ids):
        """Check that ``regexps`` filters tests created from ``ids``."""
        filtered = filters.include_regexps(regexps, create_tests_from_ids(ids))
        self.assertEqual(expected, [t.id() for t in filtered])

    def test_filter_none(self):
        self.assertFiltered(['foo', 'bar'], [], ['foo', 'bar'])

    def test_filter_one_regexp(self):
        self.assertFiltered(['foo', 'foobar', 'barfoo'], ['.*foo.*'],
                            ['foo', 'foobar', 'barfoo', 'baz'])

    def test_filter_several_regexps(self):
        self.assertFiltered(['foo', 'foobar', 'barfoo'], ['foo', 'arf'],
                            ['foo', 'foobar', 'barfoo', 'baz'])

    def test_filter_unanchored(self):
        self.assertFiltered(['foo', 'foobar', 'barfoo', 'xfoox'], ['foo'],
                            ['foo', 'foobar', 'barfoo', 'baz', 'xfoox'])

    def test_filter_start(self):
        self.assertFiltered(['foo', 'foobar'], ['^foo'],
                            ['foo', 'foobar', 'barfoo', 'baz', 'xfoox'])

    def test_filter_ends(self):
        self.assertFiltered(['foo', 'barfoo'], ['foo$'],
                            ['foo', 'foobar', 'barfoo', 'baz', 'xfoox'])


class TestFilterTestsByExcludedPrefixes(unittest.TestCase):

    def assertFiltered(self, expected, regexps, ids):
        """Check that ``prefixes`` filters tests created from ``ids``."""
        filtered = filters.exclude_regexps(regexps,
                                           create_tests_from_ids(ids))
        self.assertEqual(expected, [t.id() for t in filtered])

    def test_no_excludes(self):
        self.assertFiltered(['foo', 'bar'], [], ['foo', 'bar'])

    def test_one_exclude(self):
        self.assertFiltered(['bar'], ['foo'],
                            ['foo.bar', 'bar', 'foo.baz'])

    def test_several_excludes(self):
        self.assertFiltered(['bar'], ['foo', 'bar.'],
                            ['foo.bar', 'bar', 'foo.baz', 'bar.baz'])


class TestIncludeList(unittest.TestCase):

    def assertFiltered(self, expected, test_names, ids):
        """Check that ``test_names`` filters tests created from ``ids``."""
        filtered = filters.include_id_list(
            test_names, create_tests_from_ids(ids))
        self.assertEqual(expected, [t.id() for t in filtered])

    def test_empty(self):
        self.assertFiltered([], [], ['fail', 'pass', 'skip'])

    def test_include_knowns(self):
        self.assertFiltered(['pass', 'skip'], ['pass', 'skip'],
                            ['fail', 'pass', 'skip'])

    def test_include_with_unknowns(self):
        self.assertFiltered(['pass'],
                            ['dont.exist', 'funny.one', 'pass'],
                            ['fail', 'pass', 'skip'])

    def test_respect_list_order(self):
        self.assertFiltered(['skip', 'pass'], ['skip', 'pass'],
                            ['fail', 'pass', 'skip'])

    def test_create_duplicates(self):
        self.assertFiltered(['pass', 'pass'], ['pass', 'pass'],
                            ['fail', 'pass', 'skip'])

    def test_keep_duplicates(self):
        self.assertFiltered(['pass'], ['pass'],
                            ['fail', 'pass', 'pass'])


class TestPick(unittest.TestCase):

    def assertFiltered(self, expected, rank, modulo, ids):
        """Check that ``part`` filters tests created from ``ids``."""
        filtered = filters.pick(create_tests_from_ids(ids), rank, modulo)
        self.assertEqual(expected, [t.id() for t in filtered])

    def test_no_ratio(self):
        self.assertFiltered(['fail', 'pass', 'skip'],
                            0, 0, ['fail', 'pass', 'skip'])

    def test_first_ratio(self):
        self.assertFiltered(['fail', 'skip'],
                            1, 2, ['fail', 'pass', 'skip'])

    def test_last_ratio(self):
        self.assertFiltered(['pass'],
                            2, 2, ['fail', 'pass', 'skip'])

    def test_bogus_ratio(self):
        self.assertFiltered([],
                            12, 8,
                            ['fail', 'pass', 'skip'])


class TestPartitionSuite(unittest.TestCase):

    def assertPartition(self, expected, count, ids):
        """Check how tests created from ``ids`` are partitioned."""
        suites = filters.partition_suite(create_tests_from_ids(ids), count)
        self.assertEqual(expected, [[t.id() for t in s] for s in suites])

    def test_empty(self):
        self.assertPartition([[]], 1, [])

    def test_single(self):
        self.assertPartition([['one']], 1, ['one'])

    def test_one_by_suite(self):
        self.assertPartition([['one'], ['two']], 2, ['one', 'two'])

    def test_two_by_suite(self):
        self.assertPartition([['one', 'three'], ['two', 'four']],
                             2, ['one', 'two', 'three', 'four'])


class TestLoadTestIdList(unittest.TestCase):

    def setUp(self):
        super(TestLoadTestIdList, self).setUp()
        fixtures.set_uniq_cwd(self)

    def _create_test_list_file(self, content):
        with io.open('tests.list', 'w', encoding='utf-8') as f:
            f.write(content)
        fr = io.open('tests.list', 'r', encoding='utf-8')
        self.addCleanup(fr.close)
        return fr

    def test_load_test_list(self):
        flist = self._create_test_list_file('mod1.cl1.meth1\nmod2.cl2.meth2\n')
        tlist = filters.load_test_id_list(flist)
        assertions.assertLength(self, 2, tlist)
        self.assertEqual('mod1.cl1.meth1', tlist[0])
        self.assertEqual('mod2.cl2.meth2', tlist[1])

    def test_load_dirty_file(self):
        flist = self._create_test_list_file(
            '  mod1.cl1.meth1\n\nmod2.cl2.meth2  \nbar baz\n')
        tlist = filters.load_test_id_list(flist)
        assertions.assertLength(self, 4, tlist)
        self.assertEqual('mod1.cl1.meth1', tlist[0])
        self.assertEqual('', tlist[1])
        self.assertEqual('mod2.cl2.meth2', tlist[2])
        self.assertEqual('bar baz', tlist[3])



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Filter tests suite with regular expressions."""

import re
import unittest


def filter_suite(condition, suite):
    """Return tests for which ``condition`` is True in ``suite``.

    :param condition: A callable receiving a test and returning True if the
        test should be kept.

    :param suite: A test suite that can be iterated. It contains either tests
        or a suite inheriting from ``unittest.TestSuite``.

    ``suite`` is a tree of tests and suites, the returned suite respect the
    received suite layout, only removing empty suites.
    """
    filtered_suite = suite.__class__()
    for test in suite:
        if issubclass(test.__class__, unittest.TestSuite):
            # We received a suite, we'll filter a suite
            filtered = filter_suite(condition, test)
            if filtered.countTestCases():
                # Keep only non-empty suites
                filtered_suite.addTest(filtered)
        elif condition(test):
            # The test is kept
            filtered_suite.addTest(test)
    return filtered_suite


def pick(suite, rank, modulo):
    """Pick some tests in ``suite```.

    :param suite: The suite to pick tests from.

    :param rank: The test to pick every ``modulo`` tests when iterating
        ``suite``.

    :param modulo: How often is ``rank`` picked.

    :returns: A suite of tests picking the ``rank``nth test every ``modulo``
        tests from ``suite``.

    """
    if not modulo:
        # modulo == zero is a special case, we keep all tests
        return suite
    # global but declared inside the func so the inner 'nth' func can modify it
    global __nth_calls
    __nth_calls = 0

    def nth(test):
        global __nth_calls
        try:
            if __nth_calls + 1 == rank:
                return True
        finally:
            __nth_calls = (__nth_calls + 1) % modulo
        return False
    return filter_suite(nth, suite)


def partition_suite(suite, count):
    """Partition suite into ``count`` suites.

    :param suite: The suite to partition.

    :param count: The number of suites to build.

    :returns: A list of ``count`` suites respecting the ``suite`` layout,
        dropping inner empty suites.
    """
    # This just assigns tests in a round-robin fashion.  On one hand this
    # splits up blocks of related tests that might run faster if they share
    # resources, but on the other it avoids assigning blocks of slow tests to
    # just one partition.  So the slowest partition shouldn't be much slower
    # than the fastest (confirmed IRL, not a metric, etc ;).
    global __nth_calls  # global but declared inside the func
    suites = []
    for cur in range(count):
        __nth_calls = 0

        def nth(test):
            """Returns True for the 'cur'th suite.

            This just counts the calls and check ``cur``.
            """
            global __nth_calls
            try:
                if __nth_calls == cur:
                    return True
            finally:
                __nth_calls = (__nth_calls + 1) % count
            return False

        # Each suite is built taking each nth test, relying on filter_suite to
        # preserve the layout.
        this_runner_suite = filter_suite(nth, suite)
        suites.append(this_runner_suite)
    return suites


def iter_flat(this):
    """Iterate a suite or yield a test.

    When preserving the test suite shape (a tree in the general case) is not a
    concern, this is the easiest way to iterate a test suite.
    """
    try:
        suite = iter(this)
    except TypeError:
        # If it can't be iterated, it's a test
        yield this
    else:
        for that in suite:
            for t in iter_flat(that):
                yield t


def include_regexps(regexps, suite):
    """Returns the tests that match one of ``regexps``.

    :param regexps: A list of test id regexps (strings, will be compiled
        internally) to include. All tests are included if no regexps are
        provided.

    :param suite: The test suite to filter.
    """
    if not regexps:
        # No regexeps, no filtering
        return suite

    def matches_one_of(test):
        # A test is kept if its id matches one of the regexps
        tid = test.id()
        for reg in regexps:
            if re.search(reg, tid) is not None:
                return True
        return False
    return filter_suite(matches_one_of, suite)


def exclude_regexps(regexps, suite):
    """Returns the tests whose id does not match with any of the regexps.

    :param regexps: A list of test id regexps (strings, will be compiled
        internally) to exclude. No tests are excluded if no regexps are
        provided.

    :param suite: The test suite to filter.
    """
    if not regexps:
        # No regexeps, no filtering
        return suite

    def matches_none_of(test):
        # A test is kept if its id matches none of the regexps
        tid = test.id()
        for regexp in regexps:
            if re.search(regexp, tid):
                return False
        return True
    return filter_suite(matches_none_of, suite)


def load_test_id_list(file_handle):
    """Load a test id list from a file handle.

    The format is one test id by line.  No special care is taken to impose
    strict rules, these test ids are used to filter the test suite so a test id
    that do not match an existing test will do no harm. This allows user to add
    comments, leave blank lines, etc as they won't match existing test ids.

    On the flip side, this means no control can be made about typos (a
    non-matching test id won't be run) but in practice, the test ids are copied
    from a previous run so typos are unlikely.
    """
    test_list = []
    for test_name in file_handle.readlines():
        test_list.append(test_name.strip())
    return test_list


def include_id_list(id_list, suite):
    """Returns the tests whose id are part of the list.

    The order in the list is enforced in the returned suite. This provides a
    way to strictly control the order of execution, making isolation issues
    easier to diagnose.

    Note that this doesn't filter out duplicates in the list, instead, such
    duplicates are provided twice.

    :param id_list: A list of test ids that should be kept.

    :param suite: The test suite to filter.

    """
    if not id_list:
        # No ids, no tests
        return suite.__class__()

    filtered = suite.__class__()
    for the_id in id_list:

        def is_the_id(test):
            return test.id() == the_id
        just_the_id = filter_suite(is_the_id, suite)
        # We don't want to wrap the filtered tests into two layers of
        # suite.__class__ so we extract the single test.
        if just_the_id.countTestCases():
            # Keep only non-empty suites
            filtered.addTest(list(just_the_id)[0])
    return filtered



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Load tests from a file system tree.

Python tests themselves are generally organized as a subtree of a python
module.

There is a one to one relationship between a source file and a python module
as well as between a directory and a python package.

In the most common cases, loading all the tests for a given package is just
scanning the file system recursively from the package root, importing every
python file and loading the tests defined there if any. By default, byot will
load all the tests in packages existing in the current directory.

Alternatively, the package root can be obtained from an importable module.

Yet, some tests are not written in python and have different loading strategies
(including file/directory name matching). Even some python packages or even
just scripts may want to change this policy.

The Loader.SubLoader() method is aimed at making these policy changes as
easy as possible at any point in the python package hierarchy traversal.

Relying on the load_tests protocol, each python package or module can define
how the tests in a given subtree (down to a single module) should be loaded.
So an arbitrary tree containing tests can be populated with __init__.py files
to change the loading policy for specific subtrees. This applies equally weel
to python tests or any arbitrary kind of test (the latter requires subclassing
Loader to load the tests from the files themselves though ;).

"""
import functools
import importlib
import os
import sys
import traceback
import unittest


from byot import matchers


class Loader(unittest.TestLoader):
    """Load tests from an arbitrary tree.

    This also provides ways for packages to define the test discovery and
    loading as they see fit by subclassing.

    Sorting happens on base names inside a directory while walking the tree
    and on test class names and test method names when loading a
    module. Those sortings combined provide a test suite where test ids are
    sorted alphabetically.

    :note: The default sorting is alphabetical and is the only working
        one. While this loader sort file (and dir) names using
        'unittest.TestLoader.sortTestMethodsUsing', the later, while
        respecting that for test methods, relies on 'dir(module)' providing
        the test *classes* in alphabetical order and doesn't sort them with
        'sortTestMethodsUsing' which after all is what the name implies.
    """
    def __init__(self, *args, **kwargs):
        """Defines a test loader.

        :param files: A NameMatcher object to select the files that may contain
            tests. Default to all python source files starting with 'test'.

        :param dirs: A NameMatcher object to select the directories that may
            contain tests. Default to all directory names.

        :param root: An alternate starting point for the loading. Defaults
            to current directory.
        """
        root = kwargs.pop('root', os.getcwd())
        file_matcher = kwargs.pop(
            'file_matcher', matchers.NameMatcher(includes=[r'^test.*\.py$']))
        # MISSINGTEST: The default dir matcher should be restricted to valid
        # python identifiers. -- vila 2014-01-27
        dir_matcher = kwargs.pop(
            'dir_matcher', matchers.NameMatcher(includes=[r'^[^.]*$']))
        super(Loader, self).__init__(*args, **kwargs)
        self.root = root
        self.file_matcher = file_matcher
        self.dir_matcher = dir_matcher

    def SubLoader(self, *args, **kwargs):
        """Creates a new loader overriding load policy.

        :param root: Optional. A new root to redirect the tree traversal.

        :param file_matcher: Optional. A new matcher to filter files.

        :param dir_matcher: Optional. A new matcher to filter directories.

        :return: A new loader with the new policy, leaving the original loader
            untouched.
        """
        root = kwargs.pop('root', self.root)
        file_matcher = kwargs.pop('file_matcher', self.file_matcher)
        dir_matcher = kwargs.pop('dir_matcher', self.dir_matcher)
        # Create a clone copying the attributes that defines the current policy
        # if no specific values are provided by the caller.
        return self.__class__(*args, root=root,
                              file_matcher=file_matcher,
                              dir_matcher=dir_matcher,
                              **kwargs)

    def abspath(self, *parts):
        """Return an absolute path from a list of relative path parts.

        :param parts: Any number of path parts (including 0).

        The loader API expect relative paths but internally needs absolute
        paths from the loader root (generally the current directory when the
        loading started or the sys.path directory a package was imported from).
        """
        return os.path.join(self.root, *parts)

    def loadTestsFromTree(self, dir_path):
        """Load all tests in a tree and return the resulting suite.

        :param dir_path: The directory where the load starts.

        :return: The test suite of all collected tests in the tree, sorted, as
            defined by 'sortTestMethodsUsing', using the dir name and file
            names as keys.
        """
        suite = self.loadTestsFromPackage(dir_path)
        if suite is not None:
            return suite
        suite = self.suiteClass()
        names = os.listdir(self.abspath(dir_path))
        suite.addTests(self.loadTestsFromFiles(dir_path, names))
        return suite

    def loadTestsFromPackage(self, dir_path):
        """Load the tests defined in a python package.

        :param dir_path: The directory defining the package.

        :return: The test suite of all the tests defined in the package using
             the unittest 'load_test' protocol if appropriate. The tests are
             sorted as defined by 'sortTestMethodsUsing', using the dir name
             and file names as keys.
        """
        init_path = os.path.join(dir_path, '__init__.py')
        if not os.path.isfile(self.abspath(init_path)):
            # No python module here, nothing to load
            return None
        try:
            package = self.importFromPath(dir_path)
        except TypeError:
            # This can occur when 'dir_path' contains chars that are
            # illegal for a module name. In that case, it's not a package
            # or it should be imported from a different path (in sys.path).
            return None

        # Can we delegate to the package ?
        load_tests = getattr(package, 'load_tests', None)
        if load_tests is not None:
            # let unittest handle the 'load_tests' protocol
            return self.loadTestsFromModule(package)
        # Otherwise, If tests are defined in the package, load them

        # FIXME: Allow the suite class to be speficied in the module in a
        # load_tests compatible way (if that makes sense) -- vila 2016-04-23
        suite = self.suiteClass()
        suite.addTests(self.loadTestsFromModule(package))
        # And load the tests from the other files
        file_names = os.listdir(self.abspath(dir_path))
        file_names.remove('__init__.py')
        suite.addTests(self.loadTestsFromFiles(dir_path, file_names))
        return suite

    def loadTestsFromFiles(self, dir_path, file_names):
        """Load all tests in a list of filenames inside a dir.

        :param dir_path: The directory where the load starts.

        :param file_names: A possibly empty list of file and dir names to
            collect tests from.

        :return: The test suite of all collected tests in matched file and dir
            names in the tree rooted at 'dir_path' (the current 'file_matcher'
            and 'dir_matcher' are used against dir/file names). No sorting is
            done here but the filenames order is preserved as well as the order
            of the collected test suites.
        """
        # Walk the tree to discover the tests
        suite = self.suiteClass()
        for file_name in self.sortNames(file_names):
            rel_path = os.path.join(dir_path, file_name)
            if os.path.isfile(self.abspath(rel_path))\
                    and self.file_matcher.matches(file_name):
                suite.addTests(self.loadTestsFromFile(rel_path))
            elif os.path.isdir(self.abspath(rel_path))\
                    and self.dir_matcher.matches(file_name):
                suite.addTests(self.loadTestsFromTree(rel_path))
        return suite

    def loadTestsFromFile(self, path):
        """Load all tests in a given file.

        :param path: The path of the file.

        :return: The test suite of all tests collected in the file after the
            corresponding module is imported, using the test class and test
            method names as keys.
        """
        module = self.importFromPath(path)
        return self.loadTestsFromModule(module)

    def packageSysPathFromName(self, name):
        """Find where a package resides on disk.

        This can fail if the package cannot be imported but is guaranteed to
        succeed otherwise.

        :param name: The python package name as it appears in the import
            statement.

        :return: A (sys_path_entry, rel_path) tuple where sys_path_entry is the
            entry in sys.path from where the package is imported and rel_path
            is the relative path from that sys path entry.
        """
        mod = importlib.import_module(name)
        mod_abs_dir = os.path.abspath(os.path.dirname(mod.__file__))
        mod_rel_path = name.replace('.', os.sep)
        # Find which part of sys.path provided the module
        for p in sys.path:
            abs_path = os.path.abspath(os.path.join(p, mod_rel_path))
            if abs_path == mod_abs_dir:
                sys_path_entry = p
                break
        return sys_path_entry, mod_rel_path

    def loadTestsFromSysPathModule(self, name):
        """Load all tests in an importable module.

        :param name: The [dotted] module name.

        Instead of walking the file system from a given directory, use the
        directory the module is imported from.
        """
        sys_path_entry, mod_rel_path = self.packageSysPathFromName(name)
        subloader = self.SubLoader(root=sys_path_entry)
        return subloader.loadTestsFromPackage(mod_rel_path)

    def sortNames(self, names):
        """Return 'names' sorted as defined by sortTestMethodsUsing.

        It's a little abuse of sort*TestMethods*Using as we're sorting file
        names (or even module python paths) but it allows providing a
        consistent order for the whole suite.
        """
        return sorted(names,
                      key=functools.cmp_to_key(self.sortTestMethodsUsing))

    def importFromPath(self, path):
        imp_path = os.path.normpath(path)
        if imp_path.endswith('.py'):
            imp_path = imp_path[:-3]  # Remove the trailing '.py'
        mod_name = imp_path.replace(os.path.sep, '.')
        try:
            return importlib.import_module(mod_name)
        except ImportError:
            tb = traceback.format_exc()
            msg = 'Failed to import {} at {}:\n{}'.format(mod_name, path, tb)
            raise ImportError(msg)



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2013, 2014, 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import sys


from byot import (
    assertions,
    fixtures,
)


class TestPatch(unittest.TestCase):

    def test_new_attribute(self):
        fixtures.patch(self, self, 'attr', 'hello')
        self.assertEqual('hello', self.attr)

    def test_new_value(self):
        self.attr = 'foo'
        fixtures.patch(self, self, 'attr', 'bar')
        self.assertEqual('bar', self.attr)

    def test_preserved_value(self):
        self.attr = 'foo'

        class LocalTest(unittest.TestCase):

            # We use inner below so 'self' can still refer to the outer test
            def test_it(inner):
                fixtures.patch(inner, self, 'attr', 'bar')
                # inner sees the modified value
                inner.assertEqual('bar', self.attr)

        assertions.assertSuccessfullTest(self, LocalTest('test_it'))
        # The value has been restored
        self.assertEqual('foo', self.attr)

    def test_removed_attribute_is_restored(self):
        self.attr = 'foo'

        class LocalTest(unittest.TestCase):

            # We use inner below so 'self' can still refer to the outer test
            def test_it(inner):
                fixtures.patch(inner, self, 'attr', 'bar')
                delattr(self, 'attr')
                inner.assertEqual('not there',
                                  getattr(self, 'attr', 'not there'))

        assertions.assertSuccessfullTest(self, LocalTest('test_it'))
        # The attribute is back with the proper value
        self.assertEqual('foo', self.attr)


class TestCwdToTmp(unittest.TestCase):

    def test_dir_created_and_changed(self):
        cur = os.getcwd()
        fixtures.set_uniq_cwd(self)
        self.assertNotEqual(cur, os.getcwd())

    def test_dir_restored(self):
        self.before = os.getcwd()
        self.during = None

        # To check the cleanup we need to run an inner test
        class LocalTest(unittest.TestCase):

            # We use inner below so 'self' can still refer to the outer test
            def test_it(inner):
                # The current directory has not changed
                inner.assertEqual(self.before, os.getcwd())
                fixtures.set_uniq_cwd(inner)
                # The current directory has changed
                inner.assertNotEqual(self.before, os.getcwd())
                self.during = os.getcwd()

        assertions.assertSuccessfullTest(self, LocalTest('test_it'))
        # The current directory has been restored
        self.assertEqual(self.before, os.getcwd())

    def test_dir_default_prefix(self):
        fixtures.set_uniq_cwd(self)
        cur = os.getcwd()
        # Default is to name the tmp dir after the test method
        prefix = self._testMethodName
        self.assertTrue(os.path.basename(cur).startswith(prefix))

    def test_dir_prefix_provided(self):
        fixtures.set_uniq_cwd(self, 'provided')
        cur = os.getcwd()
        self.assertTrue(os.path.basename(cur).startswith('provided'))


class TestProtectImports(unittest.TestCase):

    def setUp(self):
        super(TestProtectImports, self).setUp()
        fixtures.protect_imports(self)

    def test_added_module_is_removed(self):
        self.assertIs(None, sys.modules.get('foo', None))

        class Test(unittest.TestCase):

            def test_it(self):
                fixtures.protect_imports(self)
                sys.modules['foo'] = 'bar'

        assertions.assertSuccessfullTest(self, Test('test_it'))
        self.assertIs(None, sys.modules.get('foo', None))

    def test_removed_module_is_restored(self):
        self.assertIs(None, sys.modules.get('I_dont_exist', None))
        sys.modules['I_dont_exist'] = 'bar'

        class Test(unittest.TestCase):

            def test_it(self):
                fixtures.protect_imports(self)
                self.assertEqual('bar', sys.modules['I_dont_exist'])
                del sys.modules['I_dont_exist']

        assertions.assertSuccessfullTest(self, Test('test_it'))
        self.assertEqual('bar', sys.modules['I_dont_exist'])

    def test_modified_module_is_restored(self):
        self.assertIs(None, sys.modules.get('I_dont_exist', None))
        sys.modules['I_dont_exist'] = 'bar'

        class Test(unittest.TestCase):

            def test_it(self):
                fixtures.protect_imports(self)
                self.assertEqual('bar', sys.modules['I_dont_exist'])
                sys.modules['I_dont_exist'] = 'qux'

        assertions.assertSuccessfullTest(self, Test('test_it'))
        self.assertEqual('bar', sys.modules['I_dont_exist'])

    def test_modified_sys_path_is_restored(self):
        fixtures.set_uniq_cwd(self)
        inserted = self.uniq_dir
        self.assertFalse(inserted in sys.path)

        class Test(unittest.TestCase):

            def test_it(self):
                fixtures.protect_imports(self)
                sys.path.insert(0, inserted)

        assertions.assertSuccessfullTest(self, Test('test_it'))
        self.assertFalse(inserted in sys.path)


class TestEnv(unittest.TestCase):

    def test_env_preserved(self):
        os.environ['NOBODY_USES_THIS'] = 'foo'

        class Inner(unittest.TestCase):

            def test_overridden(self):
                fixtures.isolate_from_env(self, {'NOBODY_USES_THIS': 'bar'})
                self.assertEqual('bar', os.environ['NOBODY_USES_THIS'])

        assertions.assertSuccessfullTest(self, Inner('test_overridden'))
        self.assertEqual('foo', os.environ['NOBODY_USES_THIS'])

    def test_env_var_deleted(self):
        os.environ['NOBODY_USES_THIS'] = 'foo'

        class Inner(unittest.TestCase):

            def test_deleted(self):
                fixtures.isolate_from_env(self, {'NOBODY_USES_THIS': None})
                self.assertIs('deleted',
                              os.environ.get('NOBODY_USES_THIS', 'deleted'))
        assertions.assertSuccessfullTest(self, Inner('test_deleted'))
        self.assertEqual('foo', os.environ['NOBODY_USES_THIS'])


class TestBuildTreeFromDesc(unittest.TestCase):

    def setUp(self):
        super(TestBuildTreeFromDesc, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_empty_description(self):
        self.assertEqual([], os.listdir('.'))
        fixtures.build_tree('')
        self.assertEqual([], os.listdir('.'))

    def test_single_line_without_return(self):
        self.assertEqual([], os.listdir('.'))
        fixtures.build_tree('file: foo')
        self.assertEqual(['foo'], os.listdir('.'))
        with open('foo') as f:
            self.assertEqual('', f.read())

    def test_leading_line_is_ignored(self):
        self.assertEqual([], os.listdir('.'))
        fixtures.build_tree('tagada\nfile: foo')
        self.assertEqual(['foo'], os.listdir('.'))
        with open('foo') as f:
            self.assertEqual('', f.read())

    def test_orphan_line_is_ignored(self):
        self.assertEqual([], os.listdir('.'))
        fixtures.build_tree('''
dir: foo
orphan line
file: foo/bar.py
baz
''')
        self.assertEqual(['foo'], os.listdir('.'))
        self.assertEqual(['bar.py'], os.listdir('foo'))
        with open('foo/bar.py') as f:
            self.assertEqual('baz\n', f.read())

    def test_empty_file_content(self):
        fixtures.build_tree('''file: foo''')
        with open('foo') as f:
            self.assertEqual('', f.read())

    def test_simple_file_content(self):
        fixtures.build_tree('''file: foo
tagada
''')
        with open('foo') as f:
            self.assertEqual('tagada\n', f.read())

    def test_file_content_in_a_dir(self):
        fixtures.build_tree('''dir: dir
file: dir/foo
bar
''')
        with open('dir/foo') as f:
            self.assertEqual('bar\n', f.read())

    def test_simple_symlink_creation(self):
        fixtures.build_tree('''file: foo
tagada
link: foo bar
''')
        with open('foo') as f:
            self.assertEqual('tagada\n', f.read())
        with open('bar') as f:  # Yeah ! Open Bar !
            self.assertEqual('tagada\n', f.read())

    def test_broken_symlink_creation(self):
        fixtures.build_tree('''link: foo bar
''')
        self.assertEqual('foo', os.readlink('bar'))

    def test_invalid_symlink_description_raises(self):
        with self.assertRaises(ValueError) as cm:
            fixtures.build_tree('''link: foo
''')
        self.assertEqual('Invalid link description: foo', cm.exception.args[0])



# This file is part of Build Your Own Tests
#
# Copyright 2018 Vincent Ladeuil
# Copyright 2014, 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from distutils import spawn

from byot import (
    assertions,
    features,
    fixtures,
)


class TestFeature(unittest.TestCase):

    def test_caching(self):

        class InstrumentedFeature(features.Feature):
            def __init__(self):
                super(InstrumentedFeature, self).__init__()
                self.calls = []

            def _probe(self):
                self.calls.append('_probe')
                return False

        feature = InstrumentedFeature()
        feature.available()
        self.assertEqual(['_probe'], feature.calls)
        # Feature._probe is called by the feature at most once.
        feature.available()
        self.assertEqual(['_probe'], feature.calls)

    def test_named(self):
        """Feature.__unicode__ should thunk to feature_name()."""
        class NamedFeature(features.Feature):
            def feature_name(self):
                return 'datsmyname'
        feature = NamedFeature()
        self.assertEqual('datsmyname', str(feature))

    def test_default(self):
        """Feature.__unicode__ should default to __class__.__name__."""
        class NamedFeature(features.Feature):
            pass
        feature = NamedFeature()
        self.assertEqual('NamedFeature', str(feature))


class TestUbuntuPlatformFeature(unittest.TestCase):

    def test_xenial(self):
        xenial = features.UbuntuPlatformFeature('xenial')
        # Inject known values
        xenial._force_release('Ubuntu', '16.04', 'xenial')
        self.assertTrue(xenial.available())


class TestAtLeastUbuntu(unittest.TestCase):

    def assertReleaseIsOlder(self, old, new):
        old_index = features.ubuntu_releases.index(old)
        new_index = features.ubuntu_releases.index(new)
        self.assertTrue(old_index < new_index)

    def test_ubuntu_releases(self):
        self.assertReleaseIsOlder('xenial', 'artful')
        self.assertReleaseIsOlder('xenial', 'bionic')
        self.assertReleaseIsOlder('artful', 'bionic')

    def test_bionic_on_xenial(self):
        at_least_bionic = features.AtLeastUbuntu('bionic')
        at_least_bionic._force_release('Ubuntu', '16.04', 'xenial')
        self.assertFalse(at_least_bionic.available())

    def test_xenial_on_bionic(self):
        at_least_xenial = features.AtLeastUbuntu('xenial')
        at_least_xenial._force_release('Ubuntu', '18.04', 'bionic')
        self.assertTrue(at_least_xenial.available())

    def test_unknown_release(self):
        at_least_unknown = features.AtLeastUbuntu('unknown')
        at_least_unknown._force_release('Ubuntu', '18.04', 'bionic')
        self.assertFalse(at_least_unknown.available())


def set_uniq_path(test):
    fixtures.set_uniq_cwd(test)
    test.bin1_dir = os.path.join(test.uniq_dir, 'bin1')
    test.bin2_dir = os.path.join(test.uniq_dir, 'bin2')
    os.mkdir(test.bin1_dir)
    os.mkdir(test.bin2_dir)
    fixtures.isolate_from_env(
        test,
        {'PATH': os.pathsep.join([test.bin1_dir, test.bin2_dir])})


def make_file(path, content, chmod_bits=0o644):
    with open(path, 'w') as f:
        f.write(content)
    os.chmod(path, chmod_bits)


class TestFindExecutable(unittest.TestCase):

    def setUp(self):
        super(TestFindExecutable, self).setUp()
        set_uniq_path(self)

    def test_not_in_path(self):
        self.assertIsNone(spawn.find_executable('I DONT EXIST'))

    def test_not_executable_is_selected_too_bad(self):
        path1 = os.path.join(self.bin1_dir, 'doit')
        make_file(path1, 'whatever')
        # Despite 'whatever' not having the 'x' chmod bit it ends up being
        # selected.
        self.assertEqual(path1, spawn.find_executable('doit'))

    def test_executable(self):
        doit_path = os.path.join(self.bin1_dir, 'doit')
        make_file(doit_path, 'whatever', 0o755)
        self.assertEqual(doit_path, spawn.find_executable('doit'))

    def test_non_executable_not_ignored(self):
        path1 = os.path.join(self.bin1_dir, 'doit')
        make_file(path1, 'whatever')
        path2 = os.path.join(self.bin2_dir, 'doit')
        make_file(path2, 'whatever', 0o755)
        # Despite 'whatever' not having the 'x' chmod bit it ends up being
        # selected.
        self.assertEqual(path1, spawn.find_executable('doit'))


class TestExecutableFeature(unittest.TestCase):

    def setUp(self):
        super(TestExecutableFeature, self).setUp()
        set_uniq_path(self)

    def test_feature_not_available(self):
        feature = features.ExecutableFeature('DO NOT EXIST')
        self.assertFalse(feature.available())

    def test_feature_available(self):
        feature = features.ExecutableFeature('doit')
        doit_path = os.path.join(self.bin1_dir, 'doit')
        make_file(doit_path, 'whatever', 0o755)
        self.assertTrue(feature.available())


class TestingFeature(features.Feature):
    """A Feature that can be set to be available or not at build time."""

    def __init__(self, available):
        super(TestingFeature, self).__init__()
        # Short-circuit the probe by directly modifying the caching attribute
        self._available = available


class TestRequires(unittest.TestCase):

    def test_skip_not_available(self):
        feature = TestingFeature(False)

        class LocalTest(unittest.TestCase):

            @features.requires(feature)
            def test_it(inner):
                raise self.failureException('This should have been skipped')

        res = assertions.assertSuccessfullTest(self, LocalTest('test_it'))
        assertions.assertLength(self, 1, res.skipped)

    def test_pass_available(self):
        feature = TestingFeature(True)

        class LocalTest(unittest.TestCase):

            @features.requires(feature)
            def test_it(inner):
                pass

        assertions.assertSuccessfullTest(self, LocalTest('test_it'))

    def test_errors_when_used_without_a_feature(self):
        feature = object()

        with self.assertRaises(AttributeError) as cm:

            class LocalTest(unittest.TestCase):

                @features.requires(feature)
                def test_it(inner):
                    msg = 'An error should have been encountered earlier'
                    raise self.failureException(msg)

        msg = "'object' object has no attribute 'available'"
        self.assertEqual(msg, str(cm.exception))


class TestTestRequires(unittest.TestCase):

    def test_skip_not_available(self):
        feature = TestingFeature(False)

        class LocalTest(unittest.TestCase):

            def test_it(inner):
                features.test_requires(inner, feature)
                raise self.failureException('This should have been skipped')

        res = assertions.assertSuccessfullTest(self, LocalTest('test_it'))
        assertions.assertLength(self, 1, res.skipped)

    def test_pass_available(self):
        feature = TestingFeature(True)

        class LocalTest(unittest.TestCase):

            def test_it(inner):
                features.test_requires(inner, feature)
                pass

        assertions.assertSuccessfullTest(self, LocalTest('test_it'))

    def test_errors_when_used_without_a_feature(self):
        feature = object()

        class LocalTest(unittest.TestCase):

            def test_it(inner):
                features.test_requires(inner, feature)
                msg = 'An error should have been encountered earlier'
                raise self.failureException(msg)

        res = unittest.TestResult()
        LocalTest('test_it').run(res)
        self.assertEqual(1, res.testsRun)
        assertions.assertLength(self, 1, res.errors)
        traceback_msg = res.errors[0][1]
        msg = "'object' object has no attribute 'available'\n"
        self.assertTrue(traceback_msg.endswith(msg))
